# -*- coding: utf-8 -*-
# package: pftk.poser.kern
from collections import namedtuple
import logging
#import traceback, sys

from langutil import RemoveExt, C_FAIL, C_OK, C_ERROR, C_FILE_NOT_FOUND, UsageError
from pypos3d.wftk.WFBasic import C_BAD_DELTAINDEX, C_BAD_DELTACOUNT, C_NODELTA, Vector3d, Point3d
from pypos3d.wftk.PoserFileParser import PoserFileParser, ParsingErrorException
from pypos3d.pftk.PoserBasic import PoserConst, PoserToken, PToken, buildRelPath, getRealPath, index, chgIndex, TAS, WBoolLine, create, Lang
from pypos3d.pftk.GeomCustom import GeomCustom
from pypos3d.pftk.StructuredAttribut import StructuredAttribut, AlternateGeom, Deltas, ValueOpDelta, ChannelMorphStatus, \
    GenericTransform, ChannelMorphStatusList, findNewDelta, filterLength, calcMapping_KDTree, CustomData, DeltaPoint


#
# Options of the report algorithm
# 
ReportOption = namedtuple('ReportOption', ['translation', \
  'maxDist', # Value changed to fit a good sphere for DAZ characters \
  'enhance',\
  'boundingType',\
  'useVicinityLoop',\
  'alpha',\
  'minVectLen'])

ReportOption.__new__.__defaults__ = ( Vector3d(), 0.025, PoserConst.C_NO_ENHANCEMENT, PoserConst.C_NO_BOUNDING, True, 0.6, 0.001)

#  
class PoserMeshedObject(StructuredAttribut):
  ''' A PoserMeshedObject represents either a Prop, an Actor, a Camera, a Light or hairs in a Poser File. '''

  def __init__(self, Name=None):
    super(PoserMeshedObject, self).__init__(Name)

    #  Figure Object (may be null) 
    self._fig = None
    self._pmName = ''
    self._hidden = False
    self._visible = True #  On
    self._bend = True #  1
  
    #  Poser V9 attribut 
    self._animatableOrigin = False
    self._dynamicsLock = True #        0
  
    #  Poser V9 attribut 
    self._collisionDetection = False
    self._addToMenu = True #     1
    self._castsShadow = True #       1
    self._includeInDepthCue = True #         1
    self._useZBuffer = True #        1
    self._parent = None
    self._conformingTarget = None
    self._creaseAngle = 80.0
  
    #  Storage class of the mesh 
    self._geomType = PoserConst.GT_NONE
  
    #  Geom Groupname when geomType is GLOBAL_OBJFILE 
    self._geomGroupName = None
  
    #  Geom filename when geomType is LOCAL_OBJFILE, formated like Poser with ':' or ""  
    self._geomFileName = None
  
    #  Base Geometry
    self._geomCustom = None
    self._geom = None
  
    #  FigureResFile with a path (relative) 
    self._FRF = None
  
    #  Channels = Ordered Dictionary of Channels (only GenericTransform) - Never null 
    #  { ChannelName in lower char : GenericTransform }
    self._channels = { }
    #  Store the group presentation of channels 
    self._groups = None
  
    #  endPoint position 
    self._endPoint = None
  
    #  origin of the mesh 
    self._origin = None
  
    #  orientation of the mesh 
    self._orientation = None
    self._displayOrigin = False
    self._displayMode = None #  USEPARENT
    self._customMaterial = 0
    self._locked = False
    self._lstAltGeom = [ ]  # For Python, never null
    self._lstWMap = [ ] # List of Weight Maps (could be implemented with a ordered dict)

    self._customData = None # Contain a Dictionnary of 'key' : custData()
  
    #  Name + Mesh definition
    self._printName = None
    self._pf = None


  def setFigureResFile(self, frf):
    self._FRF = frf

  def getFigureResFile(self): return self._FRF

  def setPoserFile(self, pf):
    self._pf = pf

  def getPoserFile(self): return self._pf

  def getChannels(self): 
    ''' Return a View on the channels '''
    return self._channels.values()

  def getGeomType(self): return self._geomType

  def setGeomType(self, gt):
    self._geomType = gt

  def getGeomGroupName(self): return self._geomGroupName

  def setGeomGroupName(self, gn):
    self._geomGroupName = gn

  def getGeomFileName(self): return self._geomFileName

  def getGeomCustom(self): return self._geomCustom

  def setGeomFileName(self, fn):
    self._geomFileName = fn

  def setFigure(self, fig):
    self._fig = fig

  def getFigure(self): return self._fig


  def getMaterial(self, matName:'str', inCustom=False):
    ''' Return the named material (None if it does not exists).
    Limit search to CustomMaterial if inCustom=True
    '''
    mat = None
    
    if self.getCustomMaterial():
      mat = next( (m for m in self.getLstAttr() \
                  if (m.getPoserType()==PoserToken.E_material) and (matName=='*' or matName==m.getName())), None)
        
    if not mat and not inCustom and self.getFigure():
      mat = self.getFigure().getMaterial(matName)
          
    return mat

  # Return the list of alternate geom (never null in Python)
  def getAltGeomList(self): return self._lstAltGeom

  #    
  def addAltGeom(self, ag=None, altGeomFile=None, poserRootDir=None, pos=-1):
    ''' Add an alternate Geometry, if it does not already exists.
    Returns
    -------
    int
      C_OK, C_FAIL if the geometrie was already existing, C_ERROR in cas of file pb
    '''
    ret = C_OK
    if not ag:
      try:
        relPath = buildRelPath(poserRootDir, altGeomFile.getCanonicalPath())

        for agold in self.getAltGeomList():
          if agold.getGeomFileName()==relPath:
            return C_FAIL

        ag = AlternateGeom()
        ag.setPoserType(PoserToken.E_alternateGeom)
        n = RemoveExt(altGeomFile.getName())
        ag.setName(n)
        ag.setPrintName(n)
        ag.setGeomFileName(relPath)

      except IOError:
        ret = C_ERROR

   
    #  Add the AlternateGeom to the specified position
    if pos>=0:
      self._lstAltGeom.insert(pos, ag)
    else:
      self._lstAltGeom.append(ag)

    #  Fix internal indexes
    self.indexAltGeom()
    return ret

  def indexAltGeom(self): #  Fix internal indexes
    for i,a in enumerate(self._lstAltGeom):
      a.setNo(i)

  # 
  # Move an alternate Geometrie to an upper position
  # @return C_OK, C_FAIL if the geometrie was already existing, 
  #         PoserConst.C_ERROR in case data inconsistency
  #    
  def moveAltGeom(self, ag):
    ret = self.removeAltGeom(ag)
    if ret == C_OK:
      ret = self.addAltGeom(ag, ag.getNo())
    return ret

  # 
  # Remove an alternate Geometrie
  # @return C_OK, C_FAIL if the alt geometrie was not existing
  #    
  def removeAltGeom(self, ag):
    res = C_OK
    try:
      self.getAltGeomList().remove(ag)
      self.indexAltGeom()
    except:
      res = C_FAIL

    return res

  # 
  # Return the alternate geom channel. Null if none
  #    
  def getAltGeomChannel(self):
    return next( (gt for gt in self.getChannels() if gt.getPoserType()==PoserToken.E_geomChan), None)


  def getChannel(self, channelName):
    return self._channels.get(channelName.lower())
  
  # Find the GenericTransform that represents a TargetGeom and that has the
  # given "channelName"
  # @param channelName
  #          Channel Name to look for.
  # @return Null if not found.
  def getTargetGeom(self, channelName):
    gt = self._channels.get(channelName.lower())
    return gt if gt and (gt.getPoserType()==PoserToken.E_targetGeom) else None

  # 
  # Find the GenericTransform that represents the given "channelName" or PoserToken
  # @param channelName
  #          Channel Name to look for.
  # @return Null if not found.
  def getGenericTransform(self, channel:'Name Or PoserToken'):
    return next( (gt for gt in self.getChannels() if gt.getPoserType()==channel), None) if isinstance(channel, PToken) \
                  else self._channels.get(channel.lower())
    

  # 
  # Update or create a targetGeom and create a dependency to the body homonyme valueParm (if bodyLinked=true)
  #
  # @param channelName
  #    *          TargetGeom name
  # @param targetFigureName
  #    *          Referenced Figure Name
  # @param srcWG
  #    *          Concerned WaveGeom
  # @param lstNewDeltas
  #    *          List of deltas
  # @param bodyIdx
  #    *          Body index of the referenced figure (silly)
  #    
  #     updateOrCreate(String channelName, String targetFigureName, WaveGeom3d srcWG, lstNewDeltas, bodyIdx, bodyLinked)
  # def updateOrCreate_0(self, channelName, targetFigureName, srcWG, lstNewDeltas, bodyIdx):
  def updateOrCreate(self, channelName, targetFigureName, srcWG, setNewDeltas, bodyIdx, bodyLinked=True):
    gt = self._channels.get(channelName.lower())
    if not gt or (gt.getPoserType()!=PoserToken.E_targetGeom): #  Create the missing targetGeom
      if bodyLinked:
        gt = GenericTransform(PoserToken.E_targetGeom, channelName, targetFigureName, \
                              self.getFigure().getRoot() if targetFigureName else None,\
                              channelName if targetFigureName else None)
                              #None if not targetFigureName else "BODY:" + str(bodyIdx),\
                              #None if not targetFigureName else channelName)
      else:
        gt = GenericTransform(PoserToken.E_targetGeom, channelName)
        #  Do not add a prefix to printable name (old P4 Stuff)
        gt.setPrintName(channelName[3:] if (channelName.startswith("PBM") or channelName.startswith("FBM")) else channelName)
        gt.addKeyFrame(0, 0.0)
        #  Default keys are static=0
        gt.setInterpStyleLocked(0)

      self.addAttribut(gt)

    gt.getDeltas().deltaSet = setNewDeltas # [ delta for delta in lstNewDeltas ] )
    gt.setNumbDeltas(len(srcWG.coordList))
    return gt

  # 
  # Update or create a valueParm and initializes it
  # @param channelName ValueParm name
  def updateOrCreateVP(self, channelName, minVal=0.0, maxVal=1.0, applyLimits=True, isHidden=False):
    vp = self._channels.get(channelName.lower())
    if not vp or (vp.getPoserType()!=PoserToken.E_valueParm):
      #  Create the missing valueParm
      vp = GenericTransform(PoserToken.E_valueParm, channelName)
      #  Do not add a prefix to printable name (old P4 Stuff)
      vp.setPrintName(channelName[3:] if (channelName.startswith("PBM") or channelName.startswith("FBM")) else channelName)
      vp.addKeyFrame(0, 0.0)
      #  Default keys are static=0
      vp.setInterpStyleLocked(0)
      self.addAttribut(vp)

    vp.setMin(minVal)
    vp.setMax(maxVal)
    vp.setForceLimits(1 if applyLimits else 0)
    vp.setHidden(isHidden)
    return vp

  

  def createDefaultChannels(self, fp:'FigurePart'):
    ''' Create the 16 minimal channels for this actor and initialize them according to the 4 textual parameters.
    Textual parameters are supposed to be 'stripped'
    
    Parameters
    ----------
    fp: FigurePart
      Description of the actor.
      Used fields are: geomGroup, translation, rotation, orientation, center
    
    Return
    ------
    int : Return Code
    
    '''
    ret = C_OK
    
    try:
      AxisOrder, vOrient = fp.getOrientation()
      
      # Get a copy of the default channels (we will modify the content)
      lstChan = PoserToken.getMinimalChannels(AxisOrder, not fp.geomGroup)
      
      # Compute Actor's center
      bc = fp.getBaryCentre()
        
      # Set the Origin and Copy it in the static channels xOffset...
      self.setOrigin(bc)
      PoserToken.setInitVect(lstChan, PoserToken.OFFSETA, bc)
      PoserToken.setInitVect(lstChan, PoserToken.OFFSETB, Point3d(bc).neg())
  
      # Set Actor's orientation
      self.setOrientation(vOrient)
      
      # Manage Translation
      tr = fp.getTranslation(self.getFigure(), bc, actor=self)
      PoserToken.setInitVect(lstChan, PoserToken.TRANS, tr)
        
      # Manage Rotation
      vri = fp.getRotation()
      PoserToken.setInitVect(lstChan, PoserToken.ROT, vri)
        
      # Create the list of channels
      for gtdef in lstChan:    
        vp = GenericTransform.createFrom(gtdef)
        self.addAttribut(vp)

    except UsageError:
      #exc_type, exc_value, exc_tb = sys.exc_info()
      #traceback.print_exception(exc_type, exc_value, exc_tb)       
      ret = fp.ret

    return ret


  def CreateShaderNodeP(self, channelName, isHidden=False):
    snp = self.updateOrCreateVP(channelName, minVal=0.0, maxVal=1.0, applyLimits=True, isHidden=isHidden)
    snp.setPoserType(PoserToken.E_shaderNodeParm)
    return snp
  
  def CreateDeformerP(self, channelName, zoneName, deformerName, isHidden=False):
    snp = self.updateOrCreateVP(channelName, minVal=-100.0, maxVal=100.0, applyLimits=True, isHidden=isHidden)
    snp.setPoserType(PoserToken.E_deformerPropChan)

    # Create Deformer Channel special attributs
    az = create(PoserToken.E_addZone, PoserToken.E_addZone.token)
    az.setValue('1 ' + zoneName)
    snp.addAttribut(az)

    dp = create(PoserToken.E_deformerProp, PoserToken.E_deformerProp.token)
    dp.setValue(deformerName)
    snp.addAttribut(dp)
    return snp
  

  def getBaseMesh(self, poserRootDir):
    if self._geom == None:
      self._geom = self.getBaseGeomCustom(poserRootDir).getWaveGeom()
    return self._geom
 
  def printBaseGeomCustom(self):
    ''' Return a printable string representing the geometry '''
    s = PoserConst.GEOMPRINT[self._geomType]
    if self._geomType==PoserConst.GT_GLOBAL_OBJFILE:
      # geomHandlerGeom 13 hip
      s = PoserConst.GEOMPRINT[self._geomType] + '[' + self._FRF.getValue() + ', ' + self._geomGroupName + ']'
    elif self._geomType==PoserConst.GT_LOCAL_OBJFILE:
      # objFileGeom 0 0 :Runtime:Geometries:Marforno:Elaya:Lamp.obj          
      s = PoserConst.GEOMPRINT[self._geomType] + '[' + self._geomFileName + ']'
    else: # Alternate should not appear here
      s = PoserConst.GEOMPRINT[self._geomType]
    return s


  def initBaseGeomCustom(self, wg:'WaveGeom'):
    ''' Initialize the internal GeomCustom for a GT_LOCAL_OBJFILE '''
    if self._geomType==PoserConst.GT_LOCAL_OBJFILE:
      self._geomCustom = GeomCustom(src=wg)
  
  def getBaseGeomCustom(self, poserRootDir):
    ''' Return the Geometry definition of the object. Without applying any morph.
    Returns
    -------
    GeomCustom
      The GeomCustom object containing the WaveGeom3d
    '''
    if not self._geomCustom:
      if (self._geomType==PoserConst.GT_NONE) or (self._geomType==PoserConst.GT_INTERNAL):
        pass
      elif self._geomType==PoserConst.GT_GLOBAL_OBJFILE:
        # geomHandlerGeom 13 hip
        gc = self._FRF.getGeomCustom(poserRootDir)
        self._geomCustom = gc.extractSortGeom(self._geomGroupName, self._geomType) if gc else None

      elif self._geomType==PoserConst.GT_LOCAL_OBJFILE:
        # objFileGeom 0 0 :Runtime:Geometries:Marforno:Elaya:Lamp.obj          
        self._geomCustom = GeomCustom(getRealPath(poserRootDir, self._geomFileName))
      else:
        logging.warning("Unknown GeomType")

    return self._geomCustom

  # TODO: Take into account new installation type (Root = Runtime)
  def getMorphedMesh(self, poserRootDir, bodyIdx, lstch):
    ''' Return a mesh morphed with the selected channels.
    Parameters
    ----------
    poserRootDir : str
      Base location of Poser Installation.
    bodyIdx : int
      Index of the figure in the poser file.
      Useless for a global prop
    lstch : List of channels.
      ChannelMorphStatusList or set of channel names
    Returns
    -------
    GeomCustom
      Return null when the object has no geometry (BODY case for example)
    '''
    resgc = None
    basegc = None
    stm = lstch.getChannelSet() if isinstance(lstch, ChannelMorphStatusList) else lstch

    # Avoid the BaseGeomCustom Extraction
    if self._geomType == PoserConst.GT_GLOBAL_OBJFILE:
      basegc = self._FRF.getGeomCustom(poserRootDir)
      copygc = GeomCustom(basegc)

      copygc.findApplyDelta(bodyIdx, self._pf, stm)
      resgc = copygc.extractSortGeom(self._geomGroupName, self._geomType)
    else:
      basegc = self.getBaseGeomCustom(poserRootDir)
      resgc = GeomCustom(basegc)
      resgc.findApplyDelta(self, self._pf, stm)
    return resgc

  def setBaseMesh(self, poserRootDir, objFileName):
    self._geomType = PoserConst.GT_LOCAL_OBJFILE
    self._geomFileName = buildRelPath(poserRootDir, objFileName)

    # 20080502 : Force Geometry reload
    self._geomCustom = None
    self._printName = None

  def getPrintName(self): return self._pmName

  def setPrintName(self, pm):
    self._pmName = pm

  def setName(self, n):
    super(PoserMeshedObject, self).setName(n)
    self._printName = None

  def getDisplayName(self):
    if self._printName == None:
      if not self._geomCustom:
        if self.getName().startswith(PoserConst.C_BODY):
          objdef = "none" if (self._FRF == None) else self._FRF.getValue()
        else:
          objdef = self._geomGroupName if self._geomType==PoserConst.GT_GLOBAL_OBJFILE else self._geomFileName
      else:
        objdef = "internal : " + str(self._geomCustom.getWaveGeom().getCoordListLength()) + " vertex"

      self._printName = self.getName() + " [" + objdef if objdef else 'None' + "]"
    return self._printName

  def isVisible(self): return self._visible

  def setVisible(self, v):
    self._visible = v

  def isHidden(self): return self._hidden

  def setHidden(self, b):
    self._hidden = b

  def isBend(self): return self._bend

  def setBend(self, b):
    self._bend = b

  def isAddToMenu(self):
    return self._addToMenu

  def setAddToMenu(self, b):
    self._addToMenu = b

  def isDisplayOrigin(self):
    return self._displayOrigin

  def setDisplayOrigin(self, b):
    self._displayOrigin = b

  def setDisplayMode(self, d:'str'):
    self._displayMode = d

  def getDisplayMode(self):
    return self._displayMode

  def getCreaseAngle(self):
    return self._creaseAngle

  def setCreaseAngle(self, d):
    self._creaseAngle = d

  def getEndPoint(self):
    return self._endPoint

  def setEndPoint(self, v):
    self._endPoint = v

  def getParent(self):
    return self._parent

  def setParent(self, p):
    self._parent = p

  def getConformingTarget(self):
    return self._conformingTarget

  def setConformingTarget(self, target):
    self._conformingTarget = target

  def getCustomMaterial(self):
    return self._customMaterial

  def setCustomMaterial(self, v):
    self._customMaterial = v

  def isLocked(self):
    return self._locked

  def getOrigin(self):
    return self._origin

  def setOrigin(self, v):
    self._origin = v

  def getOrientation(self):
    return self._orientation

  def setOrientation(self, v):
    self._orientation = v
    
  # Customer Data Management
  def setCustomData(self, key, val, withPose=0, withMat=0):
    if not self._customData:
      self._customData = CustomData()
      
    self._customData.set(key, val, withPose, withMat)

  def getCustomData(self, key=None):
    ''' Return figure's custom data.
    if key is null : return the underlying dictionnary key:custData(val, withPose, withMath)
    if key is not null : return the stored cusData
    '''
    if not self._customData:
      return None
      
    return self._customData.get(key) if key else self._customData._data

  def hasCustomData(self):
    return self._customData!=None

  def setCastsShadow(self, b):
    self._castsShadow = b

  # Create in the current MeshedObject the best deltas for the list of morph
  # targets.
  # @param poserRootDir 	To find the .OBJ files
  # @param refMeshedObj         Figure that contains initial deltas
  # @param translation          Translation between original MeshedObject and its localization in
  # this.
  # @param maxDist              Maximal distance to take a point into account. If null, no maxDist
  # are applied.
  # @param setTargetMorph       List of morphs to create
  # @param alpha                Influence of distant point (less that one). alpha = 0.6 is usually
  # a good value.
  # @return C_OK when result is correct. C_ERROR when parameters are not correct.
  def createDeltas(self, poserRootDir, refObj, setTargetMorph, ropt):
    ''' Create for the current Actor the best deltas for the list of morph targets.
    This implementation starts with an extracted refFigure for optimisation purpose.

    Parameters
    ----------
    poserRootDir : str
      Root Dir of the installation (to find the .OBJ files)

    refFigure : str
       name of the figure that contains initial deltas

    setTargetMorph : set or list of str
      Morphs' names to create
      
    ropt : ReportOption
      Tuple of options for the report algorithm

    Returns
    -------
    int
      C_OK : Success
      C_ERROR : Error in parameters
      C_FILE_NOTFOUND : Some WaveFront files were not found
    '''
    ret = C_OK
    
    if not setTargetMorph:
      return C_ERROR

    lstCurGeom = [ self.getBaseGeomCustom(poserRootDir), ]
    
    if isinstance(refObj, PoserMeshedObject):
      # Extract and Sort GeomCustom to get single numbered groups
      lstRefGeom = [ refObj.getBaseGeomCustom(poserRootDir), ]
    else: # refObj should be a Figure instance
      refPoserObject = refObj.getPoserFile()
      lstRefGeom = refPoserObject.extractAll(poserRootDir, refObj.getBodyIndex(), None)

    if not min( [ gc.isValid() for gc in lstCurGeom+lstRefGeom ] ):
      return C_FILE_NOT_FOUND

    for srcGC in lstCurGeom:
      srcWG = srcGC.getWaveGeom()

      tabMapping = calcMapping_KDTree(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)
      
      self.findReportDeltas(refObj, srcGC, srcWG, tabMapping, setTargetMorph, ropt)

    return ret


  def findReportDeltas(self, refObj, srcGC, srcWG, tabMapping, setTargetMorph, ropt, checkBODYParm=False):
    ''' Optimization function. (Would be protected in Java)
    '''
    # Foreach morph target ==> Exploit tabMapping for srcGC
    for channelName in setTargetMorph:
      logging.info("Creating deltas for %s.%s", srcGC.getName(), channelName)
      setFoundDeltas = findNewDelta(tabMapping, refObj, channelName)
      setNewDeltas = { }

      # Local Enhancement
      Deltas.enhancement(srcWG, setFoundDeltas, setNewDeltas, ropt.enhance, ropt.boundingType, ropt.useVicinityLoop, ropt.alpha)
      if len(setNewDeltas) > 0:
        # Filter the result list to remove too short vectors
        lstFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)

        self.updateOrCreate(channelName, None, srcWG, lstFinalDeltas, 0)

        if checkBODYParm:
          self.getFigure().updateOrCreateValueParm(channelName)
          
        logging.info("Deltas created for %s.%s : %d", srcGC.getName(), channelName, len(lstFinalDeltas))
      else:
        logging.warn("No Delta created for %s.%s", srcGC.getName(), channelName )



  # Create in the current meshed object the best deltas for the alternate geoms for the list of morph targets.
  # @param poserRootDir    To find the .OBJ files
  # @param refMeshedObj    Figure that contains initial deltas
  # @param ropt            Report options
  # @param setTargetMorph  List of morphs to create
  # @param otherRefObjs     A list of PoserMeshedObject where to find the morphs.
  #                        Useful when the alternate geometry has been created with some fusions
  # @return    C_OK when result is correct. C_ERROR when parameters are not correct.
  def createAlternateDeltas(self, poserRootDir, setTargetMorph, ropt, otherRefObjs=None):
    ret = C_OK
    
    if not setTargetMorph:
      return C_ERROR

    curPoserObject = self.getPoserFile()
    refMeshedObj = self

    # Find the actor      
    srcActor = curPoserObject.findActor(self.getName())

    # Extract and load alternate geoms
    lstAltG = self.getAltGeomList()
    if len(lstAltG) > 0:
      lstCurGeom = [ ]
      # Table of alternate geometry usage valueParm indicators

      # Find the geomChan Channel (should always exists if the list of alt geom is not empty)
      geomChan = self.getAltGeomChannel() # channelsAttr.findAttribut(PoserToken.E_geomChan)
      
      altno = 1
      for altg in lstAltG:
        rfn = altg.getGeomFileName()
        gc = GeomCustom(getRealPath(poserRootDir, rfn))
        lstCurGeom.append(gc)

      lstRefGeom = [ refMeshedObj.getBaseGeomCustom(poserRootDir), ]
      if otherRefObjs:        
        lstRefGeom += [ ro.getBaseGeomCustom(poserRootDir) for ro in otherRefObjs ]

      # Check File error
      if not min( [ not gc or gc.isValid() for gc in lstCurGeom+lstRefGeom ] ):
        return C_FILE_NOT_FOUND

      altno = 1
      for srcGC in lstCurGeom:
        srcWG = srcGC.getWaveGeom()

        # tabMapping = PtMapping.calcMapping(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)
        tabMapping = calcMapping_KDTree(srcWG, lstRefGeom, ropt.translation, ropt.maxDist)

        # Foreach morph target ==> Exploit tabMapping for srcGC
        for channelName in setTargetMorph:
          altChannelName = f'{channelName}_Alt{altno}'
          logging.info("Creating deltas for " + srcGC.getName() + "." + channelName + " into:" + altChannelName)
          setFoundDeltas = findNewDelta(tabMapping, refMeshedObj, channelName)
          
          # Search accros other Reference actors
          if otherRefObjs:
            for ro in otherRefObjs:
              sfd = findNewDelta(tabMapping, ro, channelName)
              setFoundDeltas.update(sfd)
            
          setNewDeltas = { }

          # Local Enhancement
          Deltas.enhancement(srcWG, setFoundDeltas, setNewDeltas, ropt.enhance, ropt.boundingType, ropt.useVicinityLoop, ropt.alpha)
          if len(setNewDeltas) > 0:
            # Filter the result list to remove too short vectors
            setFinalDeltas = filterLength(setNewDeltas, ropt.minVectLen)

            # Create or update the channel list
            destGT = self.updateOrCreate(altChannelName, self.getName(), srcWG, setFinalDeltas, srcActor.getFigure().getBodyIndex(), False)
            
            vod = ValueOpDelta(PoserToken.E_valueOpKey, srcActor.getFigure().getRefName(), srcActor.getName(), geomChan.getName(), 0.0)

            for ia in range(1, len(lstAltG)+1):
              vod.addValueKey(ia, 1.0 if (ia == altno) else 0.0)
                       
            destGT.addVOD(vod)

            # Multiply by the value of the copied channel
            vodMul = ValueOpDelta(PoserToken.E_valueOpTimes, srcActor.getFigure().getRefName(), srcActor.getName(), channelName, 0.0)
            destGT.addVOD(vodMul)

            destGT.setHidden(True)
            logging.info("Deltas created for %s.%s : %d", srcGC.getName(), channelName, len(setFinalDeltas))
        altno += 1

    return ret

  # TODO: Add logging info
  def createMorph(self, poserRootDir, morphGeomGroup, targetMorphName, masterMorphName=None, altGeomNo=0, minVectLen=0.0):
    '''Create a morph for the current PoserMeshedObject, based on an alternate geom.
    The morph applies to the base geometry or to an alternate one.
    Geometries shall have the same number of vertex (indexes matters)
    
    Parameters
    ----------
      poserRootDir : str
        Path to the poser data installation. (Usually c:\...\Poser9) 
      morphGeomGroup : GeomGroup
        GeomGroup containing the morphed (translated) vertex
      targetMorphName : str
        Name of the local Dial (channel) to activate the morph at actor/prop level
      masterMorphName : str, optional, default None
        When set, defines the name of a master dial (channel) to drive the local one.
        if it does not exist at Root (BODY) level, it will be created
      altGeomNo : int, optional, default0
        Index of the alternate geometry for the morph. If null (0), the default
        geometry of the actor/prop is chosen
      minVectLen : float, optional, default 0.0
        Minimal length of morph vectors. None if null (0.0)
    Returns
    -------
    int
      C_OK : Success
      C_FAIL : Bad index for the alternate geometry
      C_BAD_DELTAINDEX : Number of vertex differ between actor/prop and geomgroup
      C_FILE_NOTFOUND : Some Wavefront file not found
    '''
    ret = C_OK
    curPoserObject = self.getPoserFile()
    
    # Find the actor      
    srcActor = curPoserObject.findActor(self.getName())

    # Extract and load alternate geoms
    if altGeomNo>0:
      lstAltG = self.getAltGeomList()
      if altGeomNo>len(lstAltG):
        return C_FAIL
      
      # Find the geomChan Channel (should always exists if the list of alt geom is not empty)
      geomChan = self.getAltGeomChannel() # channelsAttr.findAttribut(PoserToken.E_geomChan)
      
      rfn = lstAltG[altGeomNo-1].getGeomFileName()
      realpath = getRealPath(poserRootDir, rfn)
      refGeom = GeomCustom(realpath)
      if not refGeom or not refGeom.isValid():
        logging.warning('Alternate Geometry file not found:%s for %s', realpath, self.getName())
        return C_FILE_NOT_FOUND
      
      refWG = refGeom.getWaveGeom()
      refGeomGrp = refWG.getGroups()[0]
    else: # Use base geom
      geomChan = None
      refGeom = self.getBaseGeomCustom(poserRootDir)
      if not refGeom or not refGeom.isValid():
        logging.warning('Geometry not found for:%s', self.getName())
        return C_FILE_NOT_FOUND
      refWG = refGeom.getWaveGeom()

      # Need an extract and sort to get the rigth indexes
      refGeomGrp = refWG.extractSortGeom(refWG.groups[0].getName())
      
    if len(refGeomGrp.coordList)!=len(morphGeomGroup.coordList):
      return C_BAD_DELTAINDEX

    logging.info("Creating morph for %s.%s", self.getName(), targetMorphName)

    # Create the dictionnary of non null deltas
    minsq = minVectLen*minVectLen
    setNewDeltas={}
    for nopt,vx in enumerate(refGeomGrp.coordList):
      movedVx = morphGeomGroup.coordList[nopt]
      dp = DeltaPoint(nopt, movedVx.x - vx.x, movedVx.y - vx.y, movedVx.z - vx.z)
      if dp.norme2()>minsq:
        setNewDeltas[dp.noPt] = dp

    # Create or Update the TargetGeom
    
    # Link to selected alternate geom if any
    if geomChan:      
      destGT = self.updateOrCreateVP(targetMorphName, minVal=0.0, maxVal=1.0, applyLimits=False)

      targetGT = self.updateOrCreate(f'{targetMorphName}_Alt{altGeomNo}', self.getName(), refGeomGrp, setNewDeltas, srcActor.getFigure().getBodyIndex(), False)
      
      vod = ValueOpDelta(PoserToken.E_valueOpKey, srcActor.getFigure().getRefName(), srcActor.getName(), geomChan.getName(), 0.0)

      for ia in range(1, len(lstAltG)+1):
        vod.addValueKey(ia, 1.0 if (ia == altGeomNo) else 0.0)
                       
      targetGT.addVOD(vod)
      
      # Multiply by the value of the master channel
      vodMul = ValueOpDelta(PoserToken.E_valueOpTimes, srcActor.getFigure().getRefName(), srcActor.getName(), targetMorphName, 0.0)
      targetGT.addVOD(vodMul)

      targetGT.setHidden(True)  
    else:
      destGT = self.updateOrCreate(targetMorphName, self.getName(), refGeomGrp, setNewDeltas, srcActor.getFigure().getBodyIndex(), False)    
    
    # If a masterMorphName is given: Create it at BODY level then link the destGT to it
    if masterMorphName:
      bodyAct = self.getFigure().getRootActor()
      bodyAct.updateOrCreateVP(masterMorphName, minVal=0.0, maxVal=1.0, applyLimits=False)
      vod = ValueOpDelta(PoserToken.E_valueOpDeltaAdd, srcActor.getFigure().getRefName(), bodyAct.getName(), masterMorphName, 1.0)
      destGT.addVOD(vod)
    
    logging.info("Morph created for %s.%s : %d", refGeomGrp.getName(), targetMorphName, len(setNewDeltas))

    return ret
    
  # 
  # Return the delta for a given channel and a given vertex no.
  # @param channelName
  #          Channel Name (targetGeom name in fact)
  # @param refNo
  #          Index of the vertex in the geometry of the meshed object (Actor or
  #         Prop)
  # @return Delta Object (Aka 3D vector)
  def findDelta(self, channelName, refNo):
    gt = self.getTargetGeom(channelName)
    return gt.getDeltas().deltaSet.get(refNo) if gt and gt.getDeltas() else None
          

  def deleteGTReference(self, grp, gt):
    #  FIX 2019-05-05 : Some (old) character may have not group
    if grp!=None:
      # grp._lstAttr[:] = [ po for po in grp.getLstAttr() if (po.getPoserType()!=PoserToken.E_parmNode) or (gt.getName()==po.getValue()) ]
      # grp.getLstAttr().remove(po)
      for po in grp.getLstAttr():
        if po.getPoserType()==PoserToken.E_parmNode:
          if gt.getName()==po.getValue():
            grp.getLstAttr().remove(po)
            
        elif po.getPoserType()==PoserToken.E_groupNode:
          #  Recursive, because a group may contain another group
          self.deleteGTReference(po, gt)

  def deleteChannel(self, gt):
    '''Delete the given channel.
    
    Parameter:
    ----------
    gt : str or PoserGenericTransform
      The channel to delete.    
    Returns
    -------
    int : C_OK when work done, C_FAIL if string input not found

    '''
    if isinstance(gt, str):
      # Find the channel
      gt = self._channels.get(gt.lower())
      if not gt:
        return C_FAIL
      
    del self._channels[gt.getName().lower()]
    self.deleteGTReference(self._groups, gt)
    return C_OK

  # 
  # Compute the first level list of channels that point to the given channel argument.
  # @param channel Reference channel
  # @param setIn  Set to be extended and returned. Created if null. 
  #               Does not contain duplicate channel.
  def getChannelDescendant(self, channel, setIn):
    implst = setIn if setIn else set()
    cn = channel.getName()
    gn = channel.getPoserMeshedObject().getName()
    for gt in self.getChannels():
      for vop in gt.getVOD():
        if (gn==vop.getGroupName()) and (cn==vop.getChannelName()) and (not gt in implst):
          implst.add(gt)
          break

    return implst
  
  def getWeldParent(self):
    return self.getFigure().getWeldParent(self.getName())


  # Test if the deltas are consistent with the geometry.
  # @param gt
  # @return    C_OK : No problem
  #            C_NODELTA : No Deltas
  #            C_ERROR : Geometry in read error
  def checkChannelDelta(self, gt, poserRootDir):
    res = C_OK
    if gt.ishasDeltas():
      gc = self.getBaseGeomCustom(poserRootDir)
      if gc and (gc.isValid()):
        dlt = gt.getDeltas()
        vertexSize = gc.getWaveGeom().getCoordListLength()
        if vertexSize == gt.getNumbDeltas():
          for dp in dlt.deltaSet.values():
            if dp.noPt >= vertexSize:
              return C_BAD_DELTAINDEX
        else:
          res = C_BAD_DELTACOUNT
      else:
        res = C_ERROR
    else:
      res = C_NODELTA
    return res

  # Change the name of referenced part (actor, prop, hairProp, controlProp)  
  # @param oldPartName
  # @param newPartName
  def changeReference(self, oldPartName, newPartName, changeRefToBODY=True):
    self._parent = TAS(self._parent, oldPartName, newPartName)
    self._conformingTarget = TAS(self._conformingTarget, oldPartName, newPartName)
    for gt in self.getChannels():
      gt.changeReference(oldPartName, newPartName, changeRefToBODY=changeRefToBODY)

  def setDirect(self, tokenID, val):
    if tokenID==PoserToken.E_name:
      self._pmName = val
    elif tokenID==PoserToken.E_on:
      self._visible = True
    elif tokenID==PoserToken.E_off:
      self._visible = False
    elif tokenID==PoserToken.E_hidden:
      self._hidden = val.startswith("1")
    elif tokenID==PoserToken.E_bend:
      self._bend = val.startswith("1")
    elif tokenID==PoserToken.E_animatableOrigin:
      self._animatableOrigin = val.startswith("1")
    elif tokenID==PoserToken.E_dynamicsLock:
      self._dynamicsLock = val.startswith("1")
    elif tokenID==PoserToken.E_collisionDetection:
      self._collisionDetection = val.startswith("1")
    elif tokenID==PoserToken.E_addToMenu:
      self._addToMenu = val.startswith("1")
    elif tokenID==PoserToken.E_castsShadow:
      self._castsShadow = val.startswith("1")
    elif tokenID==PoserToken.E_includeInDepthCue:
      self._includeInDepthCue = val.startswith("1")
    elif tokenID==PoserToken.E_useZBuffer:
      self._useZBuffer = val.startswith("1")
    elif tokenID==PoserToken.E_parent:
      self._parent = val
    elif tokenID==PoserToken.E_conformingTarget:
      self._conformingTarget = val
    elif tokenID==PoserToken.E_creaseAngle:
      self._creaseAngle = float(val)

    elif tokenID==PoserToken.E_geomHandlerGeom:
      self._geomType = PoserConst.GT_GLOBAL_OBJFILE
      tmps = val.strip()
      bci = tmps.rfind(' ');
      self._geomGroupName = tmps[(bci + 1):]

    elif tokenID==PoserToken.E_objFileGeom:
      self._geomType = PoserConst.GT_LOCAL_OBJFILE

      bci1 = 0

      if val.startswith(PoserConst.S_LOCAL_OBJFILE_PFX):
        bci1 = len(PoserConst.S_LOCAL_OBJFILE_PFX) -1
      else:
        bci1 = val.rfind(' ')

      self._geomFileName = val[(bci1 + 1):]

    elif tokenID==PoserToken.E_endPoint:
      self._endPoint = Vector3d.parseVector3d(val)

    elif tokenID==PoserToken.E_origin:
      self._origin = Vector3d.parseVector3d(val)

    elif tokenID==PoserToken.E_orientation:
      self._orientation = Vector3d.parseVector3d(val)

    elif tokenID==PoserToken.E_displayOrigin:
      self._displayOrigin = val.startswith("1")

    elif tokenID==PoserToken.E_locked:
      self._locked = val.startswith("1")

    elif tokenID==PoserToken.E_customMaterial:
      self._customMaterial = int(val)

    elif tokenID==PoserToken.E_displayMode:
      self._displayMode = val

    else:
      logging.warning("Unexpected Direct: (%s) %s", tokenID.token, val)

  
  # Overload add Attribut
  def addAttribut(self, po):
    if isinstance(po, GenericTransform):
      try:
        self._channels[po.getName().lower()]
        logging.warning("Duplicate Channel Definition [%s] in [%s] (overloaded)", po.getName(), self.getName())
      except KeyError:
        pass

      self._channels[po.getName().lower()] = po
      po.setPoserMeshedObject(self)
      
    elif po.getPoserType()==PoserToken.E_geomCustom:
      self._geomCustom = po
      self._geomType = PoserConst.GT_INTERNAL

    elif po.getPoserType()==PoserToken.E_storageOffset:
      # FIX20090607 : Storage Offset seen : shall not be kept in attribute list.
      return

    elif po.getPoserType()==PoserToken.E_alternateGeom:
      self._lstAltGeom.append(po)
      po.setNo(len(self._lstAltGeom))

    elif po.getPoserType()==PoserToken.E_weightMap:
      self._lstWMap.append(po)

    elif po.getPoserType()==PoserToken.E_groups:
      self._groups = po
      
    elif po.getPoserType()==PoserToken.E_readScript:
      #  TODO : Read Channel Definition Script
      logging.error("No Management rule for %s [%s]", po.getName(), po.getPoserType().token)
      
    elif po.getPoserType()==PoserToken.E_customData:
      self._customData = po

    else:
      super(PoserMeshedObject, self).addAttribut(po)

  # Specific reader to take into account Channels class removal
  #
  def read(self, st):
    inchannels = False

    while True:
      code,cn,rw = st.getLine()
      if code==PoserFileParser.TT_WORD: # Other keyword found
        try:
          vc = Lang[cn]
          #  Known word
          if vc.isStructured:
            if vc==PoserToken.E_channels:
              # Read the Openning bracket and jump to next line
              ncode,lw,rw = st.getLine()
              if ncode==PoserFileParser.TT_LEFTBACKET:
                inchannels = True
                continue
              else: # Syntax error
                logging.warning("Line[%s] - read %s instead of '{'", st.lineno(), lw)
                raise ParsingErrorException()

            sta = create(vc, rw)
            self.addAttribut(sta)
            #  Read the opening bracket
            code,cn,rw = st.getLine()
            if code==PoserFileParser.TT_LEFTBACKET:
              sta.read(st)
            else:
              if sta.isAmbi():
                st.pushBack()
              else:
                logging.warning("Line[%s]:  '{' is missing for %s", st.lineno(), cn)
                raise ParsingErrorException()
          else:
            if vc.isDirect:
              self.setDirect(vc, rw)
            else:
              sa = create(vc, cn)
              #  Read before add
              sa.read(st, rw)
              self.addAttribut(sa)
        except KeyError:  #  Mot inconnu
          logging.info("Line[%s] - Unknown word:%s", st.lineno(), cn)

      elif code==PoserFileParser.TT_RIGHTBRACKET:
        if inchannels:
          inchannels = False
        else:
          break
      elif code==PoserFileParser.TT_EOF:
        break
      else:
        logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
        raise ParsingErrorException()

    # End of read

  # Write the definition part of the meshed object.
  def writeDef(self, fw, pfx):
    nextPfx = pfx + "\t"
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n{nextPfx}{{\n")
    if self._geomType==PoserConst.GT_NONE:
      fw.write('\n')
    elif self._geomType==PoserConst.GT_GLOBAL_OBJFILE:
      fw.write(f"{nextPfx}storageOffset 0 0 0\n")
      fw.write(f"{nextPfx}geomHandlerGeom 13 {self._geomGroupName}\n")
    elif self._geomType==PoserConst.GT_LOCAL_OBJFILE:
      fw.write(f"{nextPfx}storageOffset 0 0.3487 0\n")
      fw.write(f"{nextPfx}objFileGeom 0 0 {self._geomFileName}\n")
    elif self._geomType==PoserConst.GT_INTERNAL:
      self._geomCustom.write(fw, nextPfx)
    else:
      logging.warning("Unknown GeomType at write")
    fw.write(f"{nextPfx}}}\n")

  # Write the Data part of the meshed object.
  def writeData(self, fw, pfx):
    nextPfx = pfx + "\t"
    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n{nextPfx}{{\n")
    fw.write(f"{nextPfx}name {self._pmName}\n")
    fw.write(f"{nextPfx}{'on' if self._visible else 'off'}\n")
    WBoolLine(fw, nextPfx, "bend", self._bend)
    
    if self.getPoserFile().isVersion(PoserConst.POSER_V9):
      WBoolLine(fw, nextPfx, "animatableOrigin", self._animatableOrigin)

    WBoolLine(fw, nextPfx, "dynamicsLock", self._dynamicsLock)

    if self.getPoserFile().isVersion(PoserConst.POSER_V9):
      WBoolLine(fw, nextPfx, "collisionDetection", self._collisionDetection )

    WBoolLine(fw, nextPfx, "hidden", self._hidden )
    WBoolLine(fw, nextPfx, "addToMenu", self._addToMenu )
    WBoolLine(fw, nextPfx, "castsShadow", self._castsShadow )
    WBoolLine(fw, nextPfx, "includeInDepthCue", self._includeInDepthCue )
    WBoolLine(fw, nextPfx, "useZBuffer", self._useZBuffer )

    if self._parent:
      fw.write(f"{nextPfx}parent {self._parent}\n")
    if self._conformingTarget:
      fw.write(f"{nextPfx}conformingTarget {self._conformingTarget}\n")
    if self._lstAltGeom:
      for ag in self._lstAltGeom:
        ag.write(fw, nextPfx)

    fw.write(f"{nextPfx}creaseAngle {self._creaseAngle}\n")
    
    # Print WeightMap before channels
    for po in self._lstWMap:
      po.write(fw, nextPfx)
    
    # Channels { } part
    chanPfx = nextPfx + "\t"
    fw.write(f"{nextPfx}channels\n{chanPfx}{{\n")
    if self._groups:
      self._groups.write(fw, chanPfx)
    for po in self.getChannels():
      po.write(fw, chanPfx)
    fw.write(f"{chanPfx}}}\n")

    if self._endPoint:
        fw.write(f"{nextPfx}endPoint {self._endPoint.poserPrint()}\n")
    if self._origin:
        fw.write(f"{nextPfx}origin {self._origin.poserPrint()}\n")
    if self._orientation:
        fw.write(f"{nextPfx}orientation {self._orientation.poserPrint()}\n")

    WBoolLine(fw, nextPfx, "displayOrigin", self._displayOrigin )

    if self._displayMode:
        fw.write(f"{nextPfx}displayMode {self._displayMode}\n")
    fw.write(f"{nextPfx}customMaterial {self._customMaterial}\n")
    WBoolLine(fw, nextPfx, "locked", self._locked )
    
    # Print other found attributes
    for po in self._lstAttr:
      po.write(fw, nextPfx)
    if not self._lstAttr:
      fw.write('\n')
      
    # Write Customer Data
    if self._customData:
      self._customData.write(fw, nextPfx)

    fw.write(f"{nextPfx}}}\n")

  def write(self, fw, pfx):
    raise Exception(PoserConst.C_EX_NOTCALL)




# -----------------------------------------------------------------------------
# Descendant classes of PoserMeshedObject
# -----------------------------------------------------------------------------

# 
# Just to avoid interactions with actors or props that have geoms
#  
class Camera(PoserMeshedObject):
  def __init__(self):
    super(Camera, self).__init__()

# 
# Just to avoid interactions with actors or props that have geoms
#  
class Light(PoserMeshedObject):
  def __init__(self):
    super(Light, self).__init__()

  def getLightType(self):
    sa = self.findAttribut(PoserToken.E_lightType)
    return sa.getIntegerValue()
  
  def addAttribut(self, po):
    if po.getPoserType()==PoserToken.E_depthMapSize:
      prev = self.getGenericTransform(PoserToken.E_depthMapSize)
      if prev:
        # The second defintion should be the Simple attribut
        self.getLstAttr().append(po)
        return         
      
    PoserMeshedObject.addAttribut(self, po)

# 
#  Describes a Poser "actor" object.
#  
class PoserActor(PoserMeshedObject):

  def __init__(self, Name = None):
    super(PoserActor, self).__init__(Name)
    self.idx = 0
    self.setPoserType(PoserToken.E_actor)

  #    * @see pftk.poser.kern.PoserObject#setName(java.lang.String)
  def setName(self, n):
    super(PoserActor, self).setName(n)
    self.idx = index(n)

  def getIndex(self): return self.idx

  def isBody(self):
    c = self.getName().rfind(':')
    return (c == 4) and self.getName().startswith("BODY")

  # Hide an actor.
  #   @param hidden
  #   @param menuHidden  Hide in Poser Menu
  def setHidden(self, hidden, menuHidden=None):
    self._hidden = hidden
    self._visible = False
    self._bend = False
    self._dynamicsLock = False
    self._addToMenu = menuHidden if menuHidden else not hidden
    self._castsShadow = False
    self._includeInDepthCue = False

# 
# Represents a Poser "prop" object
#  
class PoserProp(PoserMeshedObject):
  def __init__(self):
    super(PoserProp, self).__init__()

  # Props may have a figure index
  def setName(self, n):
    super(PoserProp, self).setName(n)
    self.idx = index(n)

  def getIndex(self):
    ''' Return the figure index or 0 if the prop is outside any figure ''' 
    return self.idx

  def setIndex(self, ni):
    ''' Set a new figure index for the PoserProp.
    Need to be overloaded by descendant to manage specific attributes.
    Does not change parent's index (supposed to be set externaly)
    Does not change actors referenced by channels (use changeReference for that)
    '''
    self.setName( chgIndex(self.getName(), ni) )

  def createChannelMorphList(self, stm):
    lstCS = ChannelMorphStatusList()

    #  Calculate the list of channels with regards to the known one
    for attr in self.getChannels():
      if attr.getPoserType() == PoserToken.E_targetGeom:
        lstCS.append(ChannelMorphStatus(attr, stm))

    return lstCS


class BaseProp(PoserProp):
  def __init__(self):
    super(BaseProp, self).__init__()
    self.ambiguious = True

  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    if not self._lstAttr:
      fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
    else:
      super(BaseProp, self).write(fw, pfx)


  def setIndex(self, ni):
    ''' Set a new figure index for the BaseProp. '''
    super(BaseProp, self).setIndex(ni)
    
    dp = self.getAttribut(PoserToken.E_deformerProp)
    dp.setValue( chgIndex(dp.getValue(), ni))

# 
# Represents the Magnet object of a Magner deformer.
# Just to avoid interactions with actors or props that have geoms
#  
class MagnetDeformerProp(PoserProp):
  def __init__(self):
    super(MagnetDeformerProp, self).__init__()

  def setIndex(self, ni):
    ''' Set a new figure index for the BaseProp. '''
    super(MagnetDeformerProp, self).setIndex(ni)
    
    for po in self._lstAttr:
      if po.getPoserType() in (PoserToken.E_deformTarget, PoserToken.E_addZone ):
        po.setValue( chgIndex(po.getValue(), ni))
      if po.getPoserType() == PoserToken.E_baseProp: # Remember baseProp is Ambiguous
        po.setName( chgIndex(po.getName(), ni))

# Represents the Spehirc zone of a Magner deformer.
# Just to avoid interactions with actors or props that have geoms
class SphereZoneProp(PoserProp):
  def __init__(self):
    super(SphereZoneProp, self).__init__()

# 
# Represents a Poser "path" object
#  
class CurveProp(PoserProp):
  def __init__(self):
    super(CurveProp, self).__init__()

# 
# 
# Represents a Poser "Wave" object
#  
class WaveDeformerProp(PoserProp):
  def __init__(self):
    super(WaveDeformerProp, self).__init__()


# 
# 
# Represents a Poser "Wind Field" object
#  
class ConeForceFieldProp(PoserProp):
  def __init__(self):
    super(ConeForceFieldProp, self).__init__()



# Just to avoid interactions with actors or props that have geoms
#  
class ControlProp(PoserProp):
  def __init__(self):
    super(ControlProp, self).__init__()



# 
# This class represents the hairProp attribut of a Poser File.
#  
class HairProp(PoserProp):
  def __init__(self):
    super(HairProp, self).__init__()
    self._isSimple = True
    self.ambiguious = True

    def __init__(self):
        super(HairProp, self).__init__()

  #    * @see deyme.v3d.poser.PoserObject#write(java.io.FileWriter)
  def write(self, fw, pfx):
    if self._isSimple:    
      fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")

  def addAttribut(self, po):
    super(HairProp, self).addAttribut(po)
    self._isSimple = False







# -*- coding: utf-8 -*-
'''
Comme functions and consts for Poser file analysis.
Include Poser's vocabulary up to Poser 9 (partially for Poser 11)

Created on 1 mai 2020
@author: olivier
'''
import sys, traceback
import os.path
import copy
from collections import namedtuple
import platform
import importlib

import xlrd
from xlrd.sheet import ctype_text, XL_CELL_TEXT, XL_CELL_NUMBER, XL_CELL_EMPTY

from langutil import File, C_FAIL

# Test & Set
#    * @param s
#    * @param old
#    * @param ns
#    * @return
def TAS(s, old, ns): return ns if old == s else s

def isWindows(): return platform.system()=="Windows"

# Return a relative path to the Poser root directory with 
# system file separators replaced by ":" characters.
# 
# When the fileName is not "in" the poserRootDir, fileName is returned.
# 
# @param poserRootDir
# @param fileName
# @return
def buildRelPath(poserRootDir, fileName):
  res = fileName
  if poserRootDir:
    try:
      prdcpath = os.path.realpath(poserRootDir)
      tfcpath = os.path.realpath(fileName)

      if tfcpath.startswith(prdcpath):
        res = tfcpath[len(prdcpath):]
        res = res.replace(os.sep, ':')
    except IOError:
      res = None
      print('IOError in buildRelPath(%s, %s)', poserRootDir, fileName)

  return res


#    * Build a real path name.
#    * 
#    * @param poserRootDir
#    * @param poserPath
#    * @return a matched path or null in case of inconsistency (usually with DOS drive in *nix env)
def getRealPath(poserRootDir, poserPath):
  #  From time to time relative path are enclose in double-quote " characters
  if poserPath[0]=="\"" and poserPath[-1]=="\"":
    poserPath = poserPath[1:len(poserPath)-1]

  if (len(poserPath)>3) and (poserPath[1]==':') and (poserPath[0].isalpha()):
    #  We can suppose that the name contains the DOS drive letter
    #  In such case, it's a mess on non Windows OS!
    if isWindows():
      return poserPath

    #  Try to find the last part of poserRootDir in the path
    parts = poserRootDir.split("/")
    lastPart = parts[len(parts)-1] if len(parts)>=2 else parts[1]

    lpi = poserPath.find(lastPart)
    return poserRootDir + poserPath[lpi + len(lastPart):].replace('\\', os.sep) if (lpi >= 0) else poserPath

  #  Avoid *nix absolute path
  if poserPath[0]==os.sep:
    return poserPath
  
  if poserPath.find(':') >= 0:
    poserPath = poserPath.replace(':', os.sep)
  elif poserPath.find('\\') >= 0:
    poserPath = poserPath.replace('\\', os.sep)
  
  if poserPath[0]==os.sep:
    poserPath = poserPath[1:]
  elif len(poserPath)>2 and poserPath[0]=='.' and poserPath[1]==os.sep:
    poserPath = poserPath[2:]
    
  return poserPath if os.path.isfile(poserPath) else os.path.join(poserRootDir, poserPath)
 

# Return the index of a name, i.e. the leading ":n"
# 
#   * @param src string
#   * @return an int. 0 if no ':', -1 if the input string is null
def index(n):
  if n == None:
    return -1
  c = n.rfind(':')
  return int(n[c + 1:]) if (c >= 0) else 0

def chgIndex(s,ni):
  ''' Replace a potential figure index by a new one.
  Set new one if none
  '''
  pi = s.rfind(':')
  return f'{s}:{ni}' if pi<0 else f'{s[:pi]}:{ni}'  
 
def nodeNameNo(n):
  '''
  Return the number at the end of the string.
  Return 1 if none
  '''
  for i,c in enumerate(n[::-1]):
    if not c.isdecimal(): break
  
  return int(n[-i:]) if i>0 else 1  
 
 
# Clean a name = Suppress le leading ":n" and trim (before and after)
#
# * @param src string
# * @return a cleaned string
def cleanName(src):
  src = src.strip()
  posc = src.find(':')
  return src[0:posc].strip() if (posc>=0) else src

def getPoserFileType(fn):
  pos = fn.rfind('.')
  if (pos > 0):
    ext = fn[pos+1:].lower()
    try:
      return PoserConst.PFT_EXT.index(ext)
    except ValueError:
      return C_FAIL

  return C_FAIL 

# Return true if the file type name matches a compressed form for Poser
def isCompressed(arg):
  if isinstance(arg, int):
    fileType = int(arg)
  elif isinstance(arg, str):
    fileType = getPoserFileType(arg)
  elif isinstance(arg, File):
    fileType = getPoserFileType(arg.getAbsolutePath())

  return (fileType >= 0) and (fileType < PoserConst.PFT_MAXTYPE) and ((fileType & 0x0001) == 0x0001)

def RemoveQuotes(s):
  if s and len(s)>=2:
    return s[1:-1] if ((s[0]=='"') and (s[-1]=='"')) or ((s[0]=="'") and (s[-1]=="'")) else s

  return s

def WBoolLine(fw, pfx, name, bval):
  fw.write(f"{pfx}{name} {'1' if bval else '0'}\n")

class PoserConst(object):
  C_EX_NOTCALL = 'pypos3d.pftk.error.NotCallable'
  # ========================================================== 
  # Return codes                                               
  # ========================================================== 

  # Bitmask to indicate that the figure contains an internal geometry 
  C_HAS_INTERNAL = 0x0021

  # Bitmask to indicate that the figure uses another file geometry 
  C_HAS_LOCAL_OBJFILE = 0x0022

  # Bitmask to indicate that the figure uses an alternate geometry 
  C_HAS_ALT_GEOM = 0x0024

  # File Size Differs 
  C_SIZE_CHANGED = 0x1001

  # Last Modification Date Differs 
  C_DATE_CHANGED = 0x1002

  # Bad figure index 
  C_BAD_INDEX         = -128
  C_BAD_QUALNAME      = -129

  # Actor not found in search operations 
  C_ACTOR_NOTFOUND    = -2
  C_MATERIAL_NOTFOUND = -130
  C_NODE_NOTFOUND     = -131

  # Already existing thing 
  C_EXISTS = 0x0F00

  # ========================================================== 
  # No Bounding 
  C_NO_BOUNDING = 0

  # Box Bounding 
  C_BOX_BOUNDING = 1

  # Sphere Bounding 
  C_SPHERE_BOUNDING = 2

  # Polytope Bounding 
  C_POLYTOPE_BOUNDING = 3

  # No Enhancement 
  C_NO_ENHANCEMENT = 0

  # AVG Enhancement. Based on average move of neighboors 
  C_AVG_ENHANCEMENT = 1

  # MLS Enhancement. Based on Mean Less Square calc 
  C_MLS_ENHANCEMENT = 2

  # Remove any 'BODY' const strings and replace them by a getRoot() as Figure level
  C_BODY = "BODY"

  C_UNIVERSE = "UNIVERSE"


  # ---------------------------------------------------------------------------- 
  # Token that represents links to files in a PoserFile 
  PREFVOC = [
        "readScript", # 0
        "figureResFile", # 1
        "objFileGeom", # 2
        "file", # 3 - In NodeInput, ...
        "bumpMap", # 4
        "reflectionMap", # 5
        "transparencyMap", # 6 
        "textureMap", # 7
        "morphBinaryFile", # 8 - morphBinaryFile PATH.pmd
        "injectPMDFileMorphs" # 9 - injectPMDFileMorphs JamesFBM.pmd
    ]

  PREFVOC_READSCRIPT, PREFVOC_FIGURERESFILE, PREFVOC_OBJFILEGEOM, PREFVOC_FILE,\
  PREFVOC_BUMPMAP, PREFVOC_REFLECTIONMAP, PREFVOC_TRANSPARENCYMAP, PREFVOC_TEXTUREMAP,\
  PREFVOC_MORPHBINARYFILE, PREFVOC_INJECTPMDFILEMORPHS = range(0,10)
  
  # Poser file types.<br>
  # First : uncompressed file type.
  # Next : compressed file type.
  PFT_OBJ, PFT_OBZ, \
  PFT_PZ2, PFT_P2Z, \
  PFT_PZ3, PFT_PZZ, \
  PFT_CR2, PFT_CRZ, \
  PFT_FC2, PFT_FCZ, \
  PFT_PP2, PFT_PPZ, \
  PFT_HR2, PFT_HRZ, \
  PFT_HD2, PFT_HDZ, \
  PFT_LT2, PFT_LTZ, \
  PFT_CM2, PFT_CMZ, \
  PFT_MC5, PFT_MZ5, \
  PFT_MC6, PFT_MCZ, *_ = range(0, 100)

  PFT_MAXTYPE = 0x00FF
  PFT_MASK = 0x00FE
  FFT_IMG = 0x0100
  FFT_UNKNOWN = 0xFFFF

  # Poser Language Versions names 
  POSER_V4 = "4.01"
  POSER_V5 = "5"
  POSER_V6 = "6"
  POSER_V7 = "7"
  POSER_V9 = "9"
  POSER_V11 = "11"

  # Poser file extensions.
  # ORDER shall not be changed !
  PFT_EXT = ( "obj", "obz", "pz2", "p2z", "pz3", "pzz", "cr2", "crz", "fc2", "fcz", "pp2", "ppz", "hr2", "hrz", "hd2", "hdz", "lt2", "ltz", "cm2", "cmz", "mc5", "mz5", "mc6", "mcz" )

  # Image file extensions.
  FFT_IMG_EXT = ( "jpg", "jpeg", "png", "gif", "tif", "tiff", "bmp" )

  #* Geometry definition types
  GT_NONE,GT_GLOBAL_OBJFILE,GT_LOCAL_OBJFILE,GT_INTERNAL,GT_ALTERNATE = range(5)
  GEOMTYPE = ( "None", "Main file", "Local file", "Internal", "Alternate" )
  GEOMPRINT = ( "None", "Grp", "File", "Int", "Alt" )

  # S_GLOBAL_OBJFILE_PFX = "20 "
  S_LOCAL_OBJFILE_PFX = "0 0 "
  
  C_NO_PARM = "NO_PARM"
  C_NO_NODE = "NO_NODE"
  C_NO_MAPG = '"NO_MAP"'
  C_NO_MAP = "NO_MAP"

  # Shader Node Code
  kNodeTypeCodeBLENDER  = 1000
  kNodeTypeCodeIMAGEMAP = 1001



class PoserObject(object):
  ''' Poser Object with common attributs. '''

  def __init__(self, *arg):
    ''' Create an empty PoserObject '''
    self._poserType = None
    self._name = None

  def read(self, st):
    pass

  def write(self, fw, pfx):
    pass

  # @return the name of the object.
  def getName(self): return self._name

  # Set the name.
  def setName(self, string):
    self._name = string

  #* @return Returns the poserType.
  def getPoserType(self):
    return self._poserType

  #* @param poserType The poserType to set.
  def setPoserType(self, poserType):
    self._poserType = poserType
  
  def toString(self): return self._poserType.token
        
PToken = namedtuple('PToken', ['token', 'isStructured', 'clazz', 'isDirect', 'defaultName'])
PToken.__new__.__defaults__ = ('',False,None,True,'') 

class PoserToken:
  ''' Poser class to manage Poser File language (token) and implementation modules. ''' 
  MODULES = [ 'pypos3d.pftk.SimpleAttribut', 'pypos3d.pftk.StructuredAttribut', 'pypos3d.pftk.GeomCustom', 'pypos3d.pftk.PoserMeshed', 'pypos3d.pftk.Figure' ]
  LOADED_MODULE = { }
  LOADED_CLASS = { }
  
  BADTOKEN = PToken('', False) # Illegal token
  E_addToMenu = PToken("addToMenu", False)
  E_attribute = PToken("attribute", False, None, False)
  E_bend = PToken("bend", False)
  E_animatableOrigin = PToken("animatableOrigin", False)
  E_collisionDetection = PToken("collisionDetection", False)
  E_displayOrigin = PToken("displayOrigin", False)
  E_displayMode = PToken("displayMode", False)
  E_customMaterial = PToken("customMaterial", False)
  E_uniqueInterp = PToken("uniqueInterp", False, None, False)
  E_con = PToken("con", False, None, False)
  E_lin = PToken("lin", False, None, False)
  E_br = PToken("br", False, None, False)
  E_version = PToken("version", True, 'StructuredAttribut', False)
  E_groups = PToken("groups", True, 'StructuredAttribut', False) # v5
  E_groupNode = PToken("groupNode", True, 'StructuredAttribut', False) # v 5
  E_keyLayer = PToken("keyLayer", True, 'StructuredAttribut', False) # v 5
  
  E_parmNode = PToken("parmNode", False, None, False) # v5
  E_collapsed = PToken("collapsed", False, None, False) # v4.2
  E_shaderTree = PToken("shaderTree", True, 'ShaderTree', False) # v 5
  E_node = PToken("node", True, 'Node', False) # v 5
  E_pos = PToken("pos", False, None, False) # v5
  E_showPreview = PToken("showPreview", False, None, True) # v5
  E_nodeInput = PToken("nodeInput", True, 'NodeInput', False) # v 5
  E_value = PToken("value", False) # v5
  E_parmR = PToken("parmR", False) # v5
  E_parmG = PToken("parmG", False) # v5
  E_parmB = PToken("parmB", False) # v5
  E_file = PToken("file", False) # v5
  E_backfaceCull = PToken("backfaceCull", False, None, False) # v5
  E_visibleInReflections = PToken("visibleInReflections", False, None, False) # v5
  E_visibleInRender = PToken("visibleInRender", False, None, False) # v5
  
  E_visibleInCamera = PToken("visibleInCamera", False, None, False) # v9
  E_visibleInIDL = PToken("visibleInIDL", False, None, False) # v9
  
  E_displacementBounds = PToken("displacementBounds", False, None, False) # v5
  E_shadingRate = PToken("shadingRate", False, None, False) # v5
  E_smoothPolys = PToken("smoothPolys", False, None, False) # v5
  
  E_smarParent = PToken("smartparent", False, None, False) # v5
  E_vertsGroup = PToken("vertsGroup", True, 'VertsGroup', False) # v 5
  
  E_stitchVertsGroupProperties = PToken("stitchVertsGroupProperties", True, 'StructuredAttribut', False) # v7
  E_U_Bend_Resistance = PToken("U_Bend_Resistance", False, None, False)
  E_V_Bend_Resistance = PToken("V_Bend_Resistance", False, None, False)
  E_U_Stretch_Resistance = PToken("U_Stretch_Resistance", False, None, False)
  E_V_Stretch_Resistance = PToken("V_Stretch_Resistance", False, None, False)
  E_Shear_Resistance = PToken("Shear_Resistance", False, None, False)
  E_U_Scal = PToken("U_Scale", False, None, False)
  E_V_Scal = PToken("V_Scale", False, None, False)
  E_Density = PToken("Density", False, None, False)
  E_Thickness = PToken("Thickness", False, None, False)
  
  
  E_Spring_Resistance = PToken("Spring_Resistance", False, None, False)
  E_Air_Damping = PToken("Air_Damping", False, None, False)
  E_Dynamic_Friction = PToken("Dynamic_Friction", False, None, False)
  E_Static_Friction = PToken("Static_Friction", False, None, False)
  E_Friction_Velocity_Cutoff = PToken("Friction_Velocity_Cutoff", False, None, False)
  E_Cloth_Cloth_Force = PToken("Cloth_Cloth_Force", False, None, False)
  E_U_Bend_Rate = PToken("U_Bend_Rate", False, None, False)
  E_V_Bend_Rate = PToken("V_Bend_Rate", False, None, False)
  E_Cloth_Cloth_Friction = PToken("Cloth_Cloth_Friction", False, None, False)
  E_Damping_Stretch = PToken("Damping_Stretch", False, None, False)
  E_get_friction_from_solid = PToken("get_friction_from_solid", False, None, False)
  E_Use_Edge_Springs = PToken("Use_Edge_Springs", False, None, False)
  E_anisotropic = PToken("anisotropic", False, None, False)
  
  E_shaderNodeParm = PToken("shaderNodeParm", True, 'GenericTransform', False) # v7
  
  E_rayTraceShadows = PToken("rayTraceShadows", False, None, False)
  E_depthMapShadows = PToken("depthMapShadows", False, None, False)
  E_shadowBlurRadius = PToken("shadowBlurRadius", False, None, False)
  E_shadowRaytraceSoftness = PToken("shadowRaytraceSoftness", False, None, False)
  E_shadowBiasMin = PToken("shadowBiasMin", False, None, False)
  E_doAmbientOc = PToken("doAmbientOc", False, None, False)
  E_ambientOcDist = PToken("ambientOcDist", False, None, False)
  E_ambientOcBias = PToken("ambientOcBias", False, None, False)
  E_ambientOcStrength = PToken("ambientOcStrength", False, None, False)
  E_atmosphereStrength = PToken("atmosphereStrength", False, None, False)
  
  E_userCreated = PToken("userCreated", False, None, False)
  E_toonTones = PToken("toonTones", False, None, False)
  
  E_FourPorts_p1 = PToken("FourPorts_p1", False, None, False)
  E_FourPorts_p2 = PToken("FourPorts_p2", False, None, False)
  E_FourPorts_p3 = PToken("FourPorts_p3", False, None, False)
  E_FourPorts_p4 = PToken("FourPorts_p4", False, None, False)
  
  E_ThreePortsBigLeft_p1 = PToken("ThreePortsBigLeft_p1", False, None, False)
  E_ThreePortsBigLeft_p2 = PToken("ThreePortsBigLeft_p2", False, None, False)
  E_ThreePortsBigLeft_p3 = PToken("ThreePortsBigLeft_p3", False, None, False)
  
  E_ThreePortsBigRight_p1 = PToken("ThreePortsBigRight_p1", False, None, False)
  E_ThreePortsBigRight_p2 = PToken("ThreePortsBigRight_p2", False, None, False)
  E_ThreePortsBigRight_p3 = PToken("ThreePortsBigRight_p3", False, None, False)
  
  E_ThreePortsBigTop_p1 = PToken("ThreePortsBigTop_p1", False, None, False)
  E_ThreePortsBigTop_p2 = PToken("ThreePortsBigTop_p2", False, None, False)
  E_ThreePortsBigTop_p3 = PToken("ThreePortsBigTop_p3", False, None, False)
  
  E_ThreePortsBigBottom_p1 = PToken("ThreePortsBigBottom_p1", False, None, False)
  E_ThreePortsBigBottom_p2 = PToken("ThreePortsBigBottom_p2", False, None, False)
  E_ThreePortsBigBottom_p3 = PToken("ThreePortsBigBottom_p3", False, None, False)
  
  E_TwoPortsLeftRight_p1 = PToken("TwoPortsLeftRight_p1", False, None, False)
  E_TwoPortsLeftRight_p2 = PToken("TwoPortsLeftRight_p2", False, None, False)
  
  E_TwoPortsTopBottom_p1 = PToken("TwoPortsTopBottom_p1", False, None, False)
  E_TwoPortsTopBottom_p2 = PToken("TwoPortsTopBottom_p2", False, None, False)
  
  E_aoSamples = PToken("aoSamples", False, None, False)
  E_aoDist = PToken("aoDist", False, None, False)
  E_aoBias = PToken("aoBias", False, None, False)
  E_aoAndIDL = PToken("aoAndIDL", False, None, False) # Poser 9
  E_productionFrame = PToken("productionFrame", False, None, False)
  
  E_autoScaleToView = PToken("autoScaleToView", False, None, False)
  E_resScale = PToken("resScale", False, None, False)
  E_reuseShadowMaps = PToken("reuseShadowMaps", False, None, False)
  E_useRenderer = PToken("useRenderer", False, None, False)
  E_hardwareShading = PToken("hardwareShading", False, None, False) # Poser 9
  E_hardwareShadingOptimization = PToken("hardwareShadingOptimization", False, None, False) # Poser 9
  E_hardwareShadows = PToken("hardwareShadows", False, None, False) # Poser 9
  E_previewDepthBufferOffset = PToken("previewDepthBufferOffset", False, None, False) # Poser 9
  E_previewTextureSize = PToken("previewTextureSize", False, None, False) # Poser 9
  E_previewTransLimitOn = PToken("previewTransLimitOn", False, None, False) # Poser 9
  E_transparencySort = PToken("transparencySort", False, None, False) # Poser 9
  E_previewTransLimit = PToken("previewTransLimit", False, None, False) # Poser 9
  
  E_doRealtimeAO = PToken("doRealtimeAO", False, None, False) # Poser 9
  E_realtimeAOBlurRadius = PToken("realtimeAOBlurRadius", False, None, False) # Poser 9
  E_realtimeAORes = PToken("realtimeAORes", False, None, False) # Poser 9
  E_realtimeAOStrength = PToken("realtimeAOStrength", False, None, False) # Poser 9
  E_realtimeShowBackfaces = PToken("realtimeShowBackfaces", False, None, False) # Poser 9
  E_sihouetteOutlineWidth = PToken("sihouetteOutlineWidth", False, None, False) # Poser 9
  
  E_wireframeLineEdge = PToken("wireframeLineEdge", False, None, False) # Poser 9
  E_toonEdgeLineWidth = PToken("toonEdgeLineWidth", False, None, False) # Poser 9
  
  E_giOnlyRender = PToken("giOnlyRender", False, None, False) # Poser 9
  E_giIntensity = PToken("giIntensity", False, None, False) # Poser 9
  E_giNumSamples = PToken("giNumSamples", False, None, False) # Poser 9
  E_giBounces = PToken("giBounces", False, None, False) # Poser 9
  E_giMaxError = PToken("giMaxError", False, None, False) # Poser 9
  E_giPassScale = PToken("giPassScale", False, None, False) # Poser 9
  E_useGI = PToken("useGI", False, None, False) # Poser 9
  
  E_useIrradianceCache = PToken("useIrradianceCache", False, None, False) # Poser 9
  E_hdriOutput = PToken("hdriOutput", False, None, False) # Poser 9
  E_gamma = PToken("gamma", False, None, False) # Poser 9
  E_useGamma = PToken("useGamma", False, None, False) # Poser 9
  
  E_toneMapper = PToken("toneMapper", False, None, False) # Poser 9
  E_toneGain = PToken("toneGain", False, None, False) # Poser 9
  E_toneExposure = PToken("toneExposure", False, None, False) # Poser 9
  E_subsurface = PToken("subsurface", False, None, False) # Poser 9
  
  E_settings = PToken("settings", True, 'StructuredAttribut', False)
  
  E_auto = PToken("auto", False, None, False)
  E_autoValue = PToken("autoValue", False, None, False)
  E_shadowRenderShadingRate = PToken("shadowRenderShadingRate", False, None, False)
  E_maxRayDepth = PToken("maxRayDepth", False, None, False)
  E_filterSize = PToken("filterSize", False, None, False)
  E_filterType = PToken("filterType", False, None, False)
  E_pixelSamples = PToken("pixelSamples", False, None, False)
  E_bucketSize = PToken("bucketSize", False, None, False)
  E_motionSamples = PToken("motionSamples", False, None, False)
  E_allowRayTracing = PToken("allowRayTracing", False, None, False)
  E_minShadingRate = PToken("minShadingRate", False, None, False)
  E_maxTextureRes = PToken("maxTextureRes", False, None, False)
  
  E_hairShadingRate = PToken("hairShadingRate", False, None, False)
  E_doShadows = PToken("doShadows", False, None, False)
  E_allowDisplacement = PToken("allowDisplacement", False, None, False)
  E_drawToonOutline = PToken("drawToonOutline", False, None, False)
  E_toonOutlineStyle = PToken("toonOutlineStyle", False, None, False)
  E_shadowOnlyRender = PToken("shadowOnlyRender", False, None, False)
  
  E_useP5renderer = PToken("useP5renderer", False, None, False)
  E_useDOF = PToken("useDOF", False, None, False)
  E_useSumAreaTables = PToken("useSumAreaTables", False, None, False)
  E_minDisplacementBounds = PToken("minDisplacementBounds", False, None, False)
  E_useTextureCache = PToken("useTextureCache", False, None, False)
  E_zipTextureCache = PToken("zipTextureCache", False, None, False)
  E_rayAccelerator = PToken("rayAccelerator", False, None, False)
  E_occlusionCulling = PToken("occlusionCulling", False, None, False)
  E_maxError = PToken("maxError", False, None, False)
  E_maxICSampleSize = PToken("maxICSampleSize", False, None, False)
  E_giStrength = PToken("giStrength", False, None, False)
  E_giSamples = PToken("giSamples", False, None, False)
  E_giDepth = PToken("giDepth", False, None, False)
  E_renderOptionsIndex = PToken("renderOptionsIndex", False, None, False)
  E_inputsCollapsed = PToken("inputsCollapsed", False, None, True)
  
  E_morphInfo = PToken("morphInfo", True, 'StructuredAttribut', False)
  E_morphPutty = PToken("morphPutty", True, 'StructuredAttribut', False)
  E_pin = PToken("pin", False, None, False)
  E_numWeights = PToken("numWeights", False, None, False)
  E_weights = PToken("weights", True, 'StructuredAttribut', False)
  E_w = PToken("w", False, None, False)
  E_caricatureFactor = PToken("caricatureFactor", False, None, False)
  
  E_textureInfo = PToken("textureInfo", True, 'StructuredAttribut', False)
  E_textureType = PToken("textureType", False, None, False)
  E_texturePhotoLineup = PToken("texturePhotoLineup", True, 'StructuredAttribut', False)
  
  E_frontImage = PToken("frontImage", True, 'FrontImage', False)
  E_loaded = PToken("loaded", False)
  E_cropRect = PToken("cropRect", False, None, False)
  E_numFeaturePoints = PToken("numFeaturePoints", False, None, False)
  E_featurePoints = PToken("featurePoints", True, 'StructuredAttribut', False)
  E_rotations = PToken("rotations", False, None, False)
  E_rotationsOffset = PToken("rotationsOffset", False, None, False)
  
  E_textureFile = PToken("textureFile", True, 'TextureFile', False)
  E_warped = PToken("warped", False)
  
  E_sideImage = PToken("sideImage", True, 'SideImage', False)
  
  E_bgShaderTree = PToken("bgShaderTree", True, 'StructuredAttribut', False)
  E_atmosShaderTree = PToken("atmosShaderTree", True, 'StructuredAttribut', False)
  E_faceRoom = PToken("faceRoom", True, 'StructuredAttribut', False)
  E_focusDistance = PToken("focusDistance", True, 'GenericTransform', False)
  E_fStop = PToken("fStop", True, 'GenericTransform', False)
  E_shutterOpen = PToken("shutterOpen", True, 'GenericTransform', False)
  E_shutterClose = PToken("shutterClose", True, 'GenericTransform', False)
  E_yon = PToken("yon", True, 'GenericTransform', False)
  E_rememberUndo = PToken("rememberUndo", False, None, False) # Poser 9
  
  E_depthMapStrength = PToken("depthMapStrength", True, 'GenericTransform', False)
  E_depthMapSize = PToken("depthMapSize", True, 'DepthMapSize', False)
  E_previewDepthMapSize = PToken("previewDepthMapSize", False, None, False) # Poser 9
  E_shadowSamples = PToken("shadowSamples", False, None, False) # Poser 9
  E_includeInOpenGL = PToken("includeInOpenGL", False, None, False) # Poser 9
  E_attenType = PToken("attenType", False, None, False) # Poser 9
  
  E_movieInfo = PToken("movieInfo", True, 'StructuredAttribut', False)
  E_light = PToken("light", True, 'Light', False)
  E_camera = PToken("camera", True, 'Camera', False)
  E_geomCustom = PToken("geomCustom", True, 'GeomCustom', False)
  E_alternateGeom = PToken("alternateGeom", True, 'AlternateGeom', False)
  E_geomChan = PToken("geomChan", True, 'GenericTransform', False)
  E_clothDynamicsParm = PToken("clothDynamicsParm", True, 'GenericTransform', False)
  E_channels = PToken("channels", True, 'StructuredAttribut', False)
  
  E_liteFalloffStart = PToken("liteFalloffStart", True, 'GenericTransform', False)
  E_liteFalloffEnd = PToken("liteFalloffEnd", True, 'GenericTransform', False)
  E_liteAttenStart = PToken("liteAttenStart", True, 'GenericTransform', False)
  E_liteAttenEnd = PToken("liteAttenEnd", True, 'GenericTransform', False)
  
  E_rotateX = PToken("rotateX", True, 'GenericTransform', False, 'xrot')
  E_rotateY = PToken("rotateY", True, 'GenericTransform', False, 'yrot')
  E_rotateZ = PToken("rotateZ", True, 'GenericTransform', False, 'zrot')
  E_keys = PToken("keys", True, 'Keys', False)
  E_kdIntensity = PToken("kdIntensity", True, 'GenericTransform', False)
  E_propagatingScale = PToken("propagatingScale", True, 'GenericTransform', False, 'scale')
  E_propagatingScaleX = PToken("propagatingScaleX", True, 'GenericTransform', False, 'xScale')
  E_propagatingScaleY = PToken("propagatingScaleY", True, 'GenericTransform', False, 'yScale')
  E_propagatingScaleZ = PToken("propagatingScaleZ", True, 'GenericTransform', False, 'zScale')
  E_translateX = PToken("translateX", True, 'GenericTransform', False, 'xtran')
  E_translateY = PToken("translateY", True, 'GenericTransform', False, 'ytran')
  E_translateZ = PToken("translateZ", True, 'GenericTransform', False, 'ztran')
  E_scale = PToken("scale", True, 'GenericTransform', False, 'Scale')
  E_scaleX = PToken("scaleX", True, 'Scale', False, 'xScale')
  E_scaleY = PToken("scaleY", True, 'Scale', False, 'yScale')
  E_scaleZ = PToken("scaleZ", True, 'Scale', False, 'zScale')
  E_visibility = PToken("visibility", True, 'GenericTransform', False)
  
  E_kdRed = PToken("kdRed", True, 'GenericTransform', False)
  E_kdGreen = PToken("kdGreen", True, 'GenericTransform', False)
  E_kdBlue = PToken("kdBlue", True, 'GenericTransform', False)
  E_Kd = PToken("Kd", False, None, False)
  E_Ks = PToken("Ks", False, None, False)
  E_Ns = PToken("Ns", False, None, False)
  E_antialiasing = PToken("antialiasing", False, None, False)
  E_textureStrength = PToken("textureStrength", False, None, False)
  E_objFileGeom = PToken("objFileGeom", False)
  E_objFile = PToken("objFile", False)
  E_useTexture = PToken("useTexture", False, None, False)
  E_useBump = PToken("useBump", False, None, False)
  E_castShadows = PToken("castShadows", False, None, False)
  E_renderOver = PToken("renderOver", False, None, False)
  E_toNewWindow = PToken("toNewWindow", False, None, False)
  
  E_newWinWidth = PToken("newWinWidth", False, None, False)
  E_newWinHeight = PToken("newWinHeight", False, None, False)
  E_newWinDPI = PToken("newWinDPI", False, None, False)
  E_setGeomHandlerOffset = PToken("setGeomHandlerOffset", False, None, False)
  
  E_focal = PToken("focal", True, 'GenericTransform', False)
  E_hither = PToken("hither", True, 'GenericTransform', False)
  
  E_camAutoFocal = PToken("camAutoFocal", True, 'GenericTransform', False)
  E_camAutoScale = PToken("camAutoScale", True, 'GenericTransform', False)
  E_camAutoCenterX = PToken("camAutoCenterX", True, 'GenericTransform', False)
  E_camAutoCenterY = PToken("camAutoCenterY", True, 'GenericTransform', False)
  E_camAutoCenterZ = PToken("camAutoCenterZ", True, 'GenericTransform', False)
  
  E_valueParm = PToken("valueParm", True, 'GenericTransform', False)
  E_xOffsetA = PToken("xOffsetA", True, 'GenericTransform', False, 'xOffsetA')
  E_yOffsetA = PToken("yOffsetA", True, 'GenericTransform', False, 'yOffsetA')
  E_zOffsetA = PToken("zOffsetA", True, 'GenericTransform', False, 'zOffsetA')
  E_xOffsetB = PToken("xOffsetB", True, 'GenericTransform', False, 'xTranB')
  E_yOffsetB = PToken("yOffsetB", True, 'GenericTransform', False, 'yTranB')
  E_zOffsetB = PToken("zOffsetB", True, 'GenericTransform', False, 'zTranB')
  
  E_taperX = PToken("taperX", True, 'GenericTransform', False)
  E_taperY = PToken("taperY", True, 'GenericTransform', False)
  E_taperZ = PToken("taperZ", True, 'GenericTransform', False)
  
  E_smoothScaleX = PToken("smoothScaleX", True, 'GenericTransform', False)
  E_smoothScaleY = PToken("smoothScaleY", True, 'GenericTransform', False)
  E_smoothScaleZ = PToken("smoothScaleZ", True, 'GenericTransform', False)
  E_smoothZones = PToken("smoothZones", False, None, False)
  
  E_deltas = PToken("deltas", True, 'Deltas', False)
  E_indexes = PToken("indexes", False)
  E_numbDeltas = PToken("numbDeltas", False)
  
  E_enabled = PToken("enabled", False)  # PoserV9 in targetGeom
  E_primaryParm = PToken("primaryParm", False, None, False) # PoserV9 in targetGeom
  E_useBulgeMapLeft = PToken("useBulgeMapLeft", False, None, False) # PoserV9 in targetGeom
  E_useBulgeMapRight = PToken("useBulgeMapRight", False, None, False) # PoserV9 in targetGeom
  E_usingGroups = PToken("usingGroups", False, None, False) # PoserV9 in targetGeom
  
  E_lineLength = PToken("lineLength", False, None, False)
  E_density = PToken("density", False, None, False)
  E_minRadius = PToken("minRadius", False, None, False)
  E_maxRadius = PToken("maxRadius", False, None, False)
  E_lineRandomness = PToken("lineRandomness", False, None, False)
  E_strokeHeadLength = PToken("strokeHeadLength", False, None, False)
  E_strokeTailLength = PToken("strokeTailLength", False, None, False)
  E_colorRandom = PToken("colorRandom", False, None, False)
  E_crossHatch = PToken("crossHatch", False, None, False)
  E_opacity = PToken("opacity", False, None, False)
  E_totalNormCutOff = PToken("totalNormCutOff", False, None, False)
  E_colorSegCutOff = PToken("colorSegCutOff", False, None, False)
  E_britenessSegCutOff = PToken("britenessSegCutOff", False, None, False)
  E_britenessLoSegCutOff = PToken("britenessLoSegCutOff", False, None, False)
  E_coloredLines = PToken("coloredLines", False, None, False)
  E_brushStyle = PToken("brushStyle", False, None, False)
  
  E_inkyParent = PToken("inkyParent", False, None, False)
  E_nonInkyParent = PToken("nonInkyParent", False, None, False)
  E_handGrasp = PToken("handGrasp", True, 'GenericTransform', False)
  E_thumbGrasp = PToken("thumbGrasp", True, 'GenericTransform', False)
  E_handSpread = PToken("handSpread", True, 'GenericTransform', False)
  
  E_figure = PToken("figure", True, 'FigureDescription', False)
  E_root = PToken("root", False, None, True)
  E_addChild = PToken("addChild", False, 'AddChildSA', False)
  E_skinType = PToken("skinType", False, None, False) # Poser 9
  
  E_inkyChain = PToken("inkyChain", True, 'InkyChain', False)
  E_addLink = PToken("addLink", False, None, False)
  E_goal = PToken("goal", False, None, False)
  E_linkWeight = PToken("linkWeight", False, None, False)
  E_linkParms = PToken("linkParms", False, 'LinkParmsSA', False)
  
  E_defaultPick = PToken("defaultPick", False, None, False)
  E_displayOn = PToken("displayOn", False, None, False)
  E_weld = PToken("weld", False, 'AddChildSA', False)
  
  E_doc = PToken("doc", True, 'DocDescription', False)
  E_dimensions = PToken("dimensions", False)
  E_screenPlace = PToken("screenPlace", False)
  E_useLimits = PToken("useLimits", False, None, False)
  E_headGuides = PToken("headGuides", False, None, False)
  E_horizon = PToken("horizon", False, None, False)
  E_vanishingLines = PToken("vanishingLines", False, None, False)
  E_hipShoulder = PToken("hipShoulder", False, None, False)
  E_drawBackfacing = PToken("drawBackfacing", False, None, False) # Poser 9
  E_groundDisplay = PToken("groundDisplay", False, None, False)
  E_depthCue = PToken("depthCue", False, None, False)
  E_boxesAlways = PToken("boxesAlways", False, None, False)
  E_bgPicOn = PToken("bgPicOn", False, None, False)
  E_handLockOn = PToken("handLockOn", False, None, False)
  E_loopInterpolation = PToken("loopInterpolation", False, None, False)
  E_quatInterpolation = PToken("quatInterpolation", False, None, False)
  E_doBalance = PToken("doBalance", False, None, False)
  E_fastTracking = PToken("fastTracking", False, None, False)
  E_groundShadows = PToken("groundShadows", False, None, False)
  E_bendBodies = PToken("bendBodies", False, None, False)
  E_fgColor = PToken("fgColor", False, None, False)
  E_bgColor = PToken("bgColor", False, None, False)
  E_shadowColor = PToken("shadowColor", False, None, False)
  E_paperTexture = PToken("paperTexture", False, None, False)
  E_addActor = PToken("addActor", False, None, False)
  E_pickActor = PToken("pickActor", False, None, False)
  E_addLight = PToken("addLight", False, None, False)
  E_addCamera = PToken("addCamera", False, None, False)
  E_rightCamera = PToken("rightCamera", False, None, False)
  E_leftCamera = PToken("leftCamera", False, None, False)
  E_posingCamera = PToken("posingCamera", False, None, False)
  E_faceCamera = PToken("faceCamera", False, None, False)
  E_rHandCamera = PToken("rHandCamera", False, None, False)
  E_lHandCamera = PToken("lHandCamera", False, None, False)
  E_topCamera = PToken("topCamera", False, None, False)
  E_frontCamera = PToken("frontCamera", False, None, False)
  E_mainCamera = PToken("mainCamera", False, None, False)
  E_auxCamera = PToken("auxCamera", False, None, False)
  E_dollyCamera = PToken("dollyCamera", False, None, False)
  E_useCamera = PToken("useCamera", False, None, False)
  
  E_illustrationParms = PToken("illustrationParms", True, 'StructuredAttribut', False)
  E_combineGradient = PToken("combineGradient", False, None, False)
  E_thresholdGradient = PToken("thresholdGradient", False, None, False)
  E_combineColor = PToken("combineColor", False, None, False)
  E_overBlack = PToken("overBlack", False, None, False)
  E_autoDensity = PToken("autoDensity", False, None, False)
  E_normalSegCutOff = PToken("normalSegCutOff", False, None, False)
  E_bgStyle = PToken("bgStyle", False, None, False)
  E_useUVspace = PToken("useUVspace", False, None, False)
  E_autoSpacing = PToken("autoSpacing", False, None, False)
  E_colorBlend = PToken("colorBlend", False, None, False)
  E_liteFactor1 = PToken("liteFactor1", False, None, False)
  E_liteFactor2 = PToken("liteFactor2", False, None, False)
  E_liteFactor3 = PToken("liteFactor3", False, None, False)
  E_rootActor = PToken("rootActor", False, None, False)
  
  E_fgStrokeParms = PToken("fgStrokeParms", True, 'StructuredAttribut', False)
  E_bgStrokeParms = PToken("bgStrokeParms", True, 'StructuredAttribut', False)
  E_edgeStrokeParms = PToken("edgeStrokeParms", True, 'StructuredAttribut', False)
  
  E_targetGeom = PToken("targetGeom", True, 'GenericTransform', False)
  E_pointAtParm = PToken("pointAtParm", True, 'GenericTransform', False)
  E_pointAtTarget = PToken("pointAtTarget", False, None, False)
  E_twistX = PToken("twistX", True, 'GenericTransform', False)
  E_twistY = PToken("twistY", True, 'GenericTransform', False)
  E_twistZ = PToken("twistZ", True, 'GenericTransform', False)
  E_jointX = PToken("jointX", True, 'GenericTransform', False)
  E_jointY = PToken("jointY", True, 'GenericTransform', False)
  E_jointZ = PToken("jointZ", True, 'GenericTransform', False)
  E_renderDefaults = PToken("renderDefaults", True, 'StructuredAttribut', False)
  E_curve = PToken("curve", True, 'GenericTransform', False)
  
  E_sphereZoneProp = PToken("sphereZoneProp", True, 'SphereZoneProp', False)
  E_baseProp = PToken("baseProp", True, 'BaseProp', False)
  E_magnetDeformerProp = PToken("magnetDeformerProp", True, 'MagnetDeformerProp', False)
  E_deformerPropChan = PToken("deformerPropChan", True, 'GenericTransform', False)
  E_deformerProp = PToken("deformerProp", False, None, False)
  E_deformTarget = PToken("deformTarget", False, None, False)
  E_addZone = PToken("addZone", False, None, False)
  E_newFalloffCurve = PToken("newFalloffCurve", False, None, False)
  E_falloffCpt = PToken("falloffCpt", False, None, False)
  E_buildSphere = PToken("buildSphere", False, None, False)
  E_endFalloffCurve = PToken("endFalloffCurve", False, None, False)
  E_transformPerVertex = PToken("transformPerVertex", False, None, False)
  
  E_curveProp = PToken("curveProp", True, 'CurveProp', False)
  E_numControlPts = PToken("numControlPts", False, None, False)
  E_pt = PToken("pt", False, None, False)
  E_calcCurve = PToken("calcCurve", False, None, False)
  
  E_waveDeformerProp = PToken("waveDeformerProp", True, 'WaveDeformerProp', False)
  E_wavePhase = PToken("wavePhase", True, 'GenericTransform', False)
  E_waveAmplitude = PToken("waveAmplitude", True, 'GenericTransform', False)
  E_waveLength = PToken("waveLength", True, 'GenericTransform', False)
  E_waveStretch = PToken("waveStretch", True, 'GenericTransform', False)
  E_waveAmplitudeNoise = PToken("waveAmplitudeNoise", True, 'GenericTransform', False)
  E_waveFrequency = PToken("waveFrequency", True, 'GenericTransform', False)
  E_waveSinusoidal = PToken("waveSinusoidal", True, 'GenericTransform', False)
  E_waveRectangular = PToken("waveRectangular", True, 'GenericTransform', False)
  E_waveTriangular = PToken("waveTriangular", True, 'GenericTransform', False)
  E_waveTurbulence = PToken("waveTurbulence", True, 'GenericTransform', False)
  E_waveOffset = PToken("waveOffset", True, 'GenericTransform', False)
  E_waveType = PToken("waveType", False, None, False)

  E_coneForceFieldProp = PToken("coneForceFieldProp", True, 'ConeForceFieldProp', False)
  E_forceAmplitude = PToken("forceAmplitude", True, 'GenericTransform', False)
  E_simpleFloat = PToken("simpleFloat", True, 'GenericTransform', False)
  E_subType = PToken("subType", False, None, False)


  E_otherActor = PToken("otherActor", False, None, False)
  E_matrixActor = PToken("matrixActor", False, None, False)
  E_center = PToken("center", False, None, False)
  E_startPt = PToken("startPt", False, None, False)
  E_endPt = PToken("endPt", False, None, False)
  E_flipped = PToken("flipped", False) # flipped may be followed by a boolean value.
  E_angles = PToken("angles", False, None, False)
  E_posBulgeLeft = PToken("posBulgeLeft", False, None, False)
  E_posBulgeRight = PToken("posBulgeRight", False, None, False)
  E_negBulgeLeft = PToken("negBulgeLeft", False, None, False)
  E_negBulgeRight = PToken("negBulgeRight", False, None, False)
  E_doBulge = PToken("doBulge", False, None, False)
  E_jointMult = PToken("jointMult", False, None, False)
  E_sphereMatsRaw = PToken("sphereMatsRaw", False, 'SphereMathsRaw', False)
  
  E_doShadow = PToken("doShadow", False, None, False)
  E_depthCamera = PToken("depthCamera", False, None, False)
  E_lightOn = PToken("lightOn", False, None, False)
  E_forceLimits = PToken("forceLimits", False)
  E_min = PToken("min", False)
  E_max = PToken("max", False)
  E_tMin = PToken("tMin", False, None, True)
  E_tMax = PToken("tMax", False, None, True)
  E_tExpo = PToken("tExpo", False, None, True)
  E_ksIgnoreTexture = PToken("ksIgnoreTexture", False, None, True)
  E_reflectThruLights = PToken("reflectThruLights", False, None, True)
  E_reflectThruKd = PToken("reflectThruKd", False, None, True)
  E_bumpStrength = PToken("bumpStrength", False, None, True)
  E_NsExponent = PToken("NsExponent", False, None, True)
  E_static = PToken("static", False, None, False)
  E_lightType = PToken("lightType", False, None, False)
  E_k = PToken("k", False, 'KSA', False)
  E_sl = PToken("sl", False, None, False)
  E_spl = PToken("spl", False, 'OffSA', False)
  E_valueOpDeltaAdd = PToken("valueOpDeltaAdd", False, 'ValueOpDelta', False)
  E_valueOpPlus = PToken("valueOpPlus", False, 'ValueOpDelta', False)
  E_valueOpMinus = PToken("valueOpMinus", False, 'ValueOpDelta', False)
  E_valueOpTimes = PToken("valueOpTimes", False, 'ValueOpDelta', False)
  E_valueOpDivideBy = PToken("valueOpDivideBy", False, 'ValueOpDelta', False)
  E_valueOpDivideInto = PToken("valueOpDivideInto", False, 'ValueOpDelta', False)
  E_valueOpKey = PToken("valueOpKey", False, 'ValueOpDelta', False)
  
  E_beginValueKeys = PToken("beginValueKeys", False, None, False)
  E_endValueKeys = PToken("endValueKeys", False, None, False)
  E_valueKey = PToken("valueKey", False, None, False)
  
  E_interpStyleLocked = PToken("interpStyleLocked", False) # 2015-10-30 : Converted to direct)
  E_cameraModel = PToken("cameraModel", False, None, False)
  E_staticValue = PToken("staticValue", False) # 2015-10-30 : Converted to direct )
  E_algorithm = PToken("algorithm", False, None, False)
  E_trackingScale = PToken("trackingScale", False)
  E_trackingScaleMult = PToken("trackingScaleMult", False, None, False)
  E_blendType = PToken("blendType", False, None, False)
  E_castsShadow = PToken("castsShadow", False)
  E_includeInDepthCue = PToken("includeInDepthCue", False)
  E_initValue = PToken("initValue", False)
  E_dynamicsLock = PToken("dynamicsLock", False)
  E_figureResFile = PToken("figureResFile", False, 'FigureResFile', False)
  E_hidden = PToken("hidden", False)
  E_actor = PToken("actor", True, 'PoserActor', False)
  E_controlProp = PToken("controlProp", True, 'ControlProp', False)
  E_prop = PToken("prop", True, 'PoserProp', False)
  E_material = PToken("material", True, 'PoserMaterial', False)
  E_presetMaterial = PToken("presetMaterial", True, 'PoserMaterial', False)
  E_numFrames = PToken("numFrames", False, None, False)
  E_loopStart = PToken("loopStart", False, None, False)
  E_loopEnd = PToken("loopEnd", False, None, False)
  E_currentFrame = PToken("currentFrame", False, None, False)
  E_framesPerSec = PToken("framesPerSec", False, None, False)
  E_outputWidth = PToken("outputWidth", False, None, False)
  E_outputHeight = PToken("outputHeight", False, None, False)
  E_parent = PToken("parent", False)
  E_outputFramesPerSec = PToken("outputFramesPerSec", False, None, False)
  E_outputStartFrame = PToken("outputStartFrame", False, None, False)
  E_outputEndFrame = PToken("outputEndFrame", False, None, False)
  E_useCustomFramesPerSec = PToken("useCustomFramesPerSec", False, None, False)
  E_renderStyle = PToken("renderStyle", False, None, False)
  E_outputType = PToken("outputType", False, None, False)
  E_res = PToken("res", False, None, False)
  E_movieName = PToken("movieName", False, None, False)
  E_loopMode = PToken("loopMode", False, None, False) # Poser 9
  E_bgMovieOn = PToken("bgMovieOn", False, None, False)
  E_number = PToken("number", False, None, False)
  E_name = PToken("name", False)
  E_geomResource = PToken("geomResource", False, None, False)
  E_locked = PToken("locked", False)
  
  E_0 = PToken("0", False, None, False) # Could be an error ?
  E_textureMap = PToken("textureMap", False, 'TextureMapSA', False)
  E_backgroundFile = PToken("backgroundFile", False, 'TextureMapSA', False)
  E_bumpMap = PToken("bumpMap", False, 'TextureMapSA', False)
  E_reflectionMap = PToken("reflectionMap", False, 'TextureMapSA', False)
  E_transparencyMap = PToken("transparencyMap", False, 'TextureMapSA', False)
  E_reflectionStrength = PToken("reflectionStrength", False, None, True)
  
  E_allowsBending = PToken("allowsBending", False, None, False)
  E_figureType = PToken("figureType", False, None, False)
  E_origFigureType = PToken("origFigureType", False, None, False)
  E_canonType = PToken("canonType", False, None, False)
  E_conforming = PToken("conforming", False, None, False)
  E_conformingTarget = PToken("conformingTarget", False)
  
  E_useZBuffer = PToken("useZBuffer", False)
  E_creaseAngle = PToken("creaseAngle", False)
  
  E_realTimeMode = PToken("realTimeMode", False, None, False)
  E_antiAlias = PToken("antiAlias", False, None, False)
  E_motionBlur = PToken("motionBlur", False, None, False)
  E_mBlurAmount = PToken("mBlurAmount", False, None, False)
  E_flashInfo = PToken("flashInfo", True, 'StructuredAttribut', False)
  E_numColors = PToken("numColors", False, None, False)
  E_customColors = PToken("customColors", False, None, False)
  E_color = PToken("color", False, None, False)
  
  E_quantOptions = PToken("quantOptions", False, None, False)
  E_quantFrame = PToken("quantFrame", False, None, False)
  E_overlapColors = PToken("overlapColors", False, None, False)
  E_outerLines = PToken("outerLines", False, None, False)
  E_innerLines = PToken("innerLines", False, None, False)
  E_lineWidth = PToken("lineWidth", False, None, False)
  E_autoPlay = PToken("autoPlay", False, None, False)
  
  E_storageOffset = PToken("storageOffset", False, None, False)
  E_geomHandlerGeom = PToken("geomHandlerGeom", False)
  E_endPoint = PToken("endPoint", False)
  E_origin = PToken("origin", False)
  E_orientation = PToken("orientation", False)
  E_on = PToken("on", False)
  E_off = PToken("off", False)
  E_sm = PToken("sm", False, 'OffSA', False)
  E_calcWeights = PToken("calcWeights", False, 'OffSA', False)
  E_KdColor = PToken("KdColor", False, None, True)
  E_KaColor = PToken("KaColor", False, None, True)
  E_KsColor = PToken("KsColor", False, None, True)
  E_ReflectionColor = PToken("ReflectionColor", False, None, True)
  E_TextureColor = PToken("TextureColor", False, None, True)
  
  E_port = PToken("port", False, None, False) # Use a Quad attribut
  E_d = PToken("d", False, None, False)
  E_readScript = PToken("readScript", False, None, False)
  # Use a Quad attribut   // Use a Quad attribut
  
  E_hairGrowthGroup = PToken("hairGrowthGroup", True, 'HairGrowthGroup', False)
  E_hairProp = PToken("hairProp", True, 'HairProp', False)
  E_hairDynamicsParm = PToken("hairDynamicsParm", True, 'GenericTransform', False)
  E_hairCacheFile = PToken("hairCacheFile", False, None, False)
  E_baseActor = PToken("baseActor", False, None, False)
  E_pullDown = PToken("pullDown", False, None, False)
  E_pullBack = PToken("pullBack", False, None, False)
  E_pullLeft = PToken("pullLeft", False, None, False)
  E_populationCount = PToken("populationCount", False, None, False)
  E_populationDensity = PToken("populationDensity", False, None, False)
  E_totalArea = PToken("totalArea", False, None, False)
  E_vertsPerHair = PToken("vertsPerHair", False, None, False)
  E_turbStrength = PToken("turbStrength", False, None, False)
  E_turbScale = PToken("turbScale", False, None, False)
  E_turbDelay = PToken("turbDelay", False, None, False)
  E_showPopulated = PToken("showPopulated", False, None, False)
  E_doCollisions = PToken("doCollisions", False, None, False)
  E_widthRange = PToken("widthRange", False, None, False)
  E_lengthRange = PToken("lengthRange", False, None, False)
  E_clumpiness = PToken("clumpiness", False, None, False)
  E_gravityK = PToken("gravityK", False, None, False)
  E_maxRotLenDelta = PToken("maxRotLenDelta", False, None, False)
  E_springK = PToken("springK", False, None, False)
  E_overallDamping = PToken("overallDamping", False, None, False)
  E_springDamping = PToken("springDamping", False, None, False)
  E_selfForce = PToken("selfForce", False, None, False)
  E_stiffExpo = PToken("stiffExpo", False, None, False)
  E_keepLengths = PToken("keepLengths", False, None, False)
  E_weightVertPct = PToken("weightVertPct", False, None, False)
  E_customData = PToken("customData", True, "CustomData", False)
  # PToken.E_frontImage = PoserToken("frontImage", True, FrontImage.__class__)
  # Not used Anymore E_data = PToken("data", False, None, False)
  E_weightMap = PToken("weightMap", True, "WeightMap", False) # Poser9
  E_Zones = PToken("zones", True, "Zones", False) # Poser9
  E_WeightMapZone = PToken("weightmapzone", True, "WeightMapZone", False) # Poser9
  E_sphereZone = PToken("spherezone", True, 'SphereZone', False)
  E_capsuleZone = PToken("capsulezone", True, 'CapsuleZone', False)
  E_active = PToken("active", False, None, False)
  E_math = PToken("math", False, None, False)
  E_mapname = PToken("mapname", False, None, False)
  E_outCapRadR = PToken("outCapRadR", False, None, False)
  E_outCapRadL = PToken("outCapRadL", False, None, False)
  E_outCylRadR = PToken("outCylRadR", False, None, False)
  E_outCylRadL = PToken("outCylRadL", False, None, False)
  E_inCapRadR = PToken("inCapRadR", False, None, False)
  E_inCapRadL = PToken("inCapRadL", False, None, False)
  E_inCylRadR = PToken("inCylRadR", False, None, False)
  E_inCylRadL = PToken("inCylRadL", False, None, False)
  E_outLenR = PToken("outLenR", False, None, False)
  E_outLenL = PToken("outLenL", False, None, False)
  E_inLenR = PToken("inLenR", False, None, False)
  E_inLenL = PToken("inLenL", False, None, False)
  E_matrices = PToken("matrices", False, 'SphereMathsRaw', False)
  
  # Channel initialization:
  #    PoserType, isStatic, min, max, initValue, order XYZ, ZXY, YZX, TrackingScale
  MINIMAL_CHANNELS = [ \
      # The Origin of the actor (i.e. Barycentre)
      [E_xOffsetA,   True,  -10.0, 10.0, 0.0,   0,   1,   2, 0.1], \
      [E_yOffsetA,   True,  -10.0, 10.0, 0.0,   1,   2,   0, 0.1], \
      [E_zOffsetA,   True,  -10.0, 10.0, 0.0,   2,   0,   1, 0.1], \

      [E_scale,      False,  0.1, 100.0, 1.0,   3,   3,   3, 0.01], \
      [E_scaleX,     False,  0.1, 100.0, 1.0,   4,   5,   6, 0.01], \
      [E_scaleY,     False,  0.1, 100.0, 1.0,   5,   6,   4, 0.01], \
      [E_scaleZ,     False,  0.1, 100.0, 1.0,   6,   4,   5, 0.01], \
      
      [E_rotateX,    False, -180.0,  180.0, 0.0,   7,   8,   9, 1.0], \
      [E_rotateY,    False, -180.0,  180.0, 0.0,   8,   9,   7, 1.0], \
      [E_rotateZ,    False, -180.0,  180.0, 0.0,   9,   7,   8, 1.0], \

      # The oppositte vector to Barycentre
      [E_xOffsetB,   True,  -10.0, 10.0, 0.0,  10,  11,  12, 0.1], \
      [E_yOffsetB,   True,  -10.0, 10.0, 0.0,  11,  12,  10, 0.1], \
      [E_zOffsetB,   True,  -10.0, 10.0, 0.0,  12,  10,  11, 0.1], \
      
      [E_translateX, False, -10.0, 10.0, 0.0,  13,  14,  15, 0.01], \
      [E_translateY, False, -10.0, 10.0, 0.0,  14,  15,  13, 0.01], \
      [E_translateZ, False, -10.0, 10.0, 0.0,  15,  13,  14, 0.01], \
      ]

  # Channel initialization:
  #    PoserType, isStatic, min, max, initValue,       order XYZ, ZXY, YZX
  MINIMAL_BODY_CHANNELS = [ \
      # The Origin of the actor (i.e. Barycentre)
      [E_xOffsetA,   True,  -1000.0, 1000.0,  0.0,   0,   1,   2, 0.1], \
      [E_yOffsetA,   True,  -1000.0, 1000.0, 0.74,   1,   2,   0, 0.1], \
      [E_zOffsetA,   True,  -1000.0, 1000.0,  0.0,   2,   0,   1, 0.1], \

      [E_propagatingScale,      False,     0.1, 1000.0, 1.0,   3,   3,   3, 0.01], \
      [E_propagatingScaleX,     False,     0.1, 1000.0, 1.0,   4,   5,   6, 0.01], \
      [E_propagatingScaleY,     False,     0.1, 1000.0, 1.0,   5,   6,   4, 0.01], \
      [E_propagatingScaleZ,     False,     0.1, 1000.0, 1.0,   6,   4,   5, 0.01], \
      
      [E_rotateX,    False,  -180.0,  180.0, 0.0,   7,   8,   9, 1.0], \
      [E_rotateY,    False,  -180.0,  180.0, 0.0,   8,   9,   7, 1.0], \
      [E_rotateZ,    False,  -180.0,  180.0, 0.0,   9,   7,   8, 1.0], \

      # The oppositte vector to Barycentre
      [E_xOffsetB,   True,  -1000.0, 1000.0, 0.0,  10,  11,  12, 0.1], \
      [E_yOffsetB,   True,  -1000.0, 1000.0, 0.0,  11,  12,  10, 0.1], \
      [E_zOffsetB,   True,  -1000.0, 1000.0, 0.0,  12,  10,  11, 0.1], \
      
      [E_translateX, False, -1000.0, 1000.0, 0.0,  13,  14,  15, 0.01], \
      [E_translateY, False, -1000.0, 1000.0, 0.0,  14,  15,  13, 0.01], \
      [E_translateZ, False, -1000.0, 1000.0, 0.0,  15,  13,  14, 0.01], \
      ]
  
  ORENTATIONS = { 'xyz':5, 'yzx':7, 'zxy':6 }
  
  SCALES  = (E_scale, E_scaleX, E_scaleY, E_scaleZ,\
            E_propagatingScale, E_propagatingScaleX, E_propagatingScaleY, E_propagatingScaleZ)
  
  ROT     = (E_rotateX, E_rotateY, E_rotateZ)
  TRANS   = (E_translateX, E_translateY, E_translateZ)
  OFFSETA = (E_xOffsetA, E_yOffsetA, E_zOffsetA)
  OFFSETB = (E_xOffsetB, E_yOffsetB, E_zOffsetB)

  @classmethod
  def setInitVect(clz, lst, props, v):
    ch = next(t for t in lst if t[0]==props[0])
    ch[4] = v.x
    ch = next(t for t in lst if t[0]==props[1])
    ch[4] = v.y
    ch = next(t for t in lst if t[0]==props[2])
    ch[4] = v.z

  @classmethod
  def getMinimalChannels(cls, axisOrder='xyz', isBody=False):
    '''
    Return the list of ordered channels to create either for an actor or 
    for the Root actor (usually BODY)
    '''
    # Get a copy of the default channels (we will modify the content)
    lstChan = copy.deepcopy(PoserToken.MINIMAL_BODY_CHANNELS if isBody else PoserToken.MINIMAL_CHANNELS )
    # Sort the Channel table according to actor orientation
    col = PoserToken.ORENTATIONS[axisOrder]
    lstChan.sort(key=lambda t: t[col])
    return lstChan


Lang = { getattr(PoserToken, ex).token : getattr(PoserToken, ex) for ex in dir(PoserToken) if ex[0:2]=='E_' }
LangList = [ getattr(PoserToken, ex) for ex in dir(PoserToken) if ex[0:2]=='E_' ]
LowerLangName = { getattr(PoserToken, ex):getattr(PoserToken, ex).token.lower()  for ex in dir(PoserToken) if ex[0:2]=='E_' }
LowerChanTypes = { getattr(PoserToken, ex).token.lower():getattr(PoserToken, ex)  for ex in dir(PoserToken) if ex[0:2]=='E_' and getattr(PoserToken, ex).clazz=='GenericTransform' }

# In Python: Implemented by a module level function in the highest 
# Module of this object model
def create(ptk, name):
  ''' Create a Structured or a Simple Attribut '''
  class_ = None
  classNameToLoad = ptk.clazz
  
  try:
    if not classNameToLoad:
      if ptk.isStructured:
        # Illegal case should not append
        raise(Exception("Token({0:s}) - null Class for a Structured".format(ptk.token)))      
      else:
        classNameToLoad = 'SimpleAttribut'
      
    class_ = PoserToken.LOADED_CLASS[classNameToLoad]
      
  except KeyError:
    # print("Token({0:s}) - Class to be loaded".format(ptk.token, ptk.clazz if ptk.clazz else ''))
    
    for modulename in PoserToken.MODULES:
      try:
        module = PoserToken.LOADED_MODULE[modulename]
      except KeyError:
        # Useless now: print("Loading:" + modulename)
        module = importlib.import_module(modulename)
        PoserToken.LOADED_MODULE.update( { modulename : module })
      
      try:
        class_ = getattr(module, classNameToLoad)
        PoserToken.LOADED_CLASS.update( { classNameToLoad : class_ })
        break
      except AttributeError:
        #print('Not the right module - Attribut Error on [{0:s}]:{1:s}'.format(ptk.token, ptk.clazz))
        pass
  
  if class_:    
    try:
      sa = class_()    
      sa.setName(name)
      sa.setPoserType(ptk)
    except:
      #exc_type, exc_value, exc_traceback = sys.exc_info() #ioex.printStackTrace()
      print('Instanciantion Error on [{0:s}]:{1:s}'.format(ptk.token, ptk.clazz if ptk.clazz else 'default' ))
      #print(repr(traceback.extract_tb(exc_traceback)))
      traceback.print_last()
  else:
    print('No module for [{0:s}]:{1:s}'.format(ptk.token, ptk.clazz))
    sa = None
      
  return sa


#  private static void readXLSLine(HSSFRow row, String[][] desttab, int rowid, int nbcols)
def __readXLSLine(sheet, desttab, rowid, nbcols):
  desttab.append([ ])

  for c in range(0, nbcols):
    value = None
    cell = sheet.cell(rowid, c)
    if not cell:
      print("CELL col(c)=" + str(c) + " VALUE=NULL")
    else:
      if cell.ctype==XL_CELL_TEXT:
        value = cell.value
      elif cell.ctype==XL_CELL_NUMBER:
        value = str(cell.value)
      elif cell.ctype==XL_CELL_EMPTY:
        value = None
      else:
        #XL_CELL_DATE:'xldate', 
        #XL_CELL_BOOLEAN:'bool', 
        #XL_CELL_ERROR:'error', 
        #XL_CELL_BLANK:'blank'}
        cell_type_str = ctype_text.get(cell.ctype, 'unknown type')
        print('Not Managed({0:d},{1:d}) : {2:s}'.format(rowid,c,cell_type_str))

    desttab[rowid].append(value)

# XL (xls) file reader. This method reads the first sheet and converts it into 
#  a string table
#  public static String[][] readXLSFile(String filename)
def readXLSFile(filename):
  book = xlrd.open_workbook(filename)
 
  if (not book) or (book.nsheets==0):
    return None

  # Just import the first sheet
  k = 0
  #HSSFSheet sheet = wb.getSheetAt(k);
  sheet = book.sheet_by_index(k)
  rows = sheet.nrows
  print("Sheet " + str(k) + " \"" + book.sheet_names()[k] + "\" has " + str(rows) + " row(s).")

  tabres = [ ] # new String[rows][];

  # The number of columns of the first row should be the size of the table
  row0 = sheet.row_values(0)
  nbcols = len(row0) #row0.getPhysicalNumberOfCells();

  __readXLSLine(sheet, tabres, 0, nbcols)

  for r in range(1, rows):
    row = sheet.row_values(r)
    if row:
      __readXLSLine(sheet, tabres, r, nbcols)

  return tabres
  

if __name__ == '__main__':
  for tk in LangList[1:]:
      attr = create(tk, 'toto')
      print('Object({0:s}) size={1:d}'.format(tk.token, sys.getsizeof(attr)))
      try:
        attr.write(sys.stdout,'')
      except Exception as e:
        if e.args[0]==PoserConst.C_EX_NOTCALL:
          attr.writeDef(sys.stdout, '')
          attr.writeData(sys.stdout,'')
        else:
          traceback.print_last()
        
        









# -*- coding: utf-8 -*-
'''
  pypos3d.wftk intends to be a highlevel library for:
  - WaveFront Files (3D geometry files)

  Requires : numpy, scipy.spatial
'''
__version__ = '1.5.0'
__author__  = 'Olivier DUFAILLY'
__license__ = 'BSD'



'''
Created on 23 nov. 2020

  pypos3d.propslim implements a decimate algorithm
  Requires : pypos3d.wftk, numpy

  Decimation function.
  This package 'propslim' contains the port of the QSlim algorithm initially 
  developed and distributed by Michael Garland in July 2004 within the "SlimKit
  Surface Modeling Tools".
  This decimate function applies to "WaveGeom" and preserves 'texture' coordinates.

  **WARNING**
  Original copyrights and credentials are delivered within this independant module.
  The original C++ code was GPL2. I don't know about my partial port, I suspect it to be _copylefted_ by it.
  If it's not the case, I would choose for 'New BSD license'
  For commercial usage in a closed source program, I suggest to remove the package: pypos3d.propslim

@author: olivier
'''
import logging

from langutil import C_FAIL
from pypos3d.wftk.WFBasic import Point3d, TexCoord2f, C_FACE_TEXT, FEPSILON3
from pypos3d.wftk.WaveGeom import WaveGeom
from pypos3d.propslim import MX_UNBOUND, MX_ALL_MASK, MX_VALID_FLAG, MX_PERVERTEX, MX_MAX_BINDING


class MxFace: # implements Blockable
  # Enhancement: Used sloted class for performance aspect
  v0:int
  v1:int
  v2:int
  mark:int
  tag:int
  grpNo:int # Original groupe id
  matIdx:int # Preserve original material index

  def __init__(self, v0, v1, v2, matIdx, grpNo):
    ''' Constructors: -1 supposed not used for real vertex indexes
        NOMORE:  public MxFace()
        public MxFace(/* MxVertexID */int v0, /* MxVertexID */int v1, /* MxVertexID */int v2)
        NOMORE:  public MxFace(final MxFace f)
    '''
    self.v0 = v0
    self.v1 = v1
    self.v2 = v2
    self.mark = 0
    self.tag = MX_VALID_FLAG
    self.grpNo = grpNo # Original groupe id
    self.matIdx = matIdx # Preserve original material index

  def remap_vertex(self, ifrom:'MxVertexID', ito:'MxVertexID'):
    nmapped = 0
    if self.v0 == ifrom:
      self.v0 = ito
      nmapped+=1
    if self.v1 == ifrom:
      self.v1 = ito
      nmapped+=1
    if self.v2 == ifrom:
      self.v2 = ito
      nmapped+=1

    return nmapped

  def get(self, ptno): return self.v0 if ptno==0 else self.v1 if ptno==1 else self.v2

  def __str__(self):
    return 'F[{:d},{:d},{:d}, mat={:d}]'.format(self.v0, self.v1, self.v2, self.matIdx)


# 
class MxStdModel(object):
  '''
  public class MxStdModel extends MxBlockModel implements MixKit
  MxBlockModel inlined
  '''

  # public int binding_mask
  def __init__(self):
    self.vertices = [ ] # MxDynBlock(nvert, Point3d()) # MxVertex())
    self.protvert = set() # Set of protected vertices
    self.lstMat = [ 'default' ]
    self.faces = [ ] #MxDynBlock(nface, MxFace())
    self.tcoords = None
    self.tbinding = MX_UNBOUND
    self.binding_mask = MX_ALL_MASK
    #self.v_data = MxDynBlock(nvert, vertex_data()) --> v_data will be hold in Point3d (new attributes)
    #self.f_data = MxDynBlock(nface, face_data()) --> f_data are hold in MxFace
  
    # Replaced by a list(list(int))
    self.face_links = [] # MxDynBloc(nvert, MxDynBlock(6, int()), True)

  def add_vertex(self, x, y, z):
    np = Point3d(x, y, z)
    np.mark = 0
    np.tag = MX_VALID_FLAG
    self.vertices.append(np)
    self.face_links.append([])
    l = len(self.face_links)-1 # .last_id()
    return l
    
  #public/* MxFaceID */int add_face(int v1, int v2, int v3, boolean will_link /* =True */)
  def add_face(self, v1, v2, v3, isProtected, matIdx, grpNo):
    # FIX: Check Triangle surface before adding it
    # Compute each edges length
    cp0,cp1,cp2 = self.vertices[v1], self.vertices[v2], self.vertices[v3]
    l0,l1,l2 = cp0.distance(cp1), cp1.distance(cp2), cp2.distance(cp0)
      
    # Compute Surface (Heron formula)
    # Protect from rounding effects that create very small negative s2
    p = 0.5*(l0+l1+l2)
    s2 = p*(p-l0)*(p-l1)*(p-l2) # s2 is the Squared surface of the triangle
    if s2<FEPSILON3: # Rejected Triangle      
      logging.warning("Triangle[%d, %d, %d] rejected: too small (%g)", v1, v2, v3, s2)
      return C_FAIL
    else:
      f = MxFace(v1, v2, v3, matIdx, grpNo)
      self.faces.append(f)
      vid = len(self.faces)-1
  
      # We always link if will_link: # self.init_face(vid)
      self.face_links[v1].append(vid)
      self.face_links[v2].append(vid)
      self.face_links[v3].append(vid)
  
      if isProtected:
        self.protvert.add(v1)
        self.protvert.add(v2)
        self.protvert.add(v3)
      
    return vid

  def add_texcoord(self, s, t):
    self.tcoords.append(TexCoord2f(s, t))
    return len(self.tcoords)-1 # .last_id()

  # Never called by PropSlim Algo
  def remove_vertex(self, v): 
    del self.face_links[v]
    del self.vertices[v]
    
    if self.texcoord_binding() == MX_PERVERTEX:
      del self.tcoords[v]


  @classmethod
  def binding_size(cls, m, i:'byte'):
    if i==MX_UNBOUND:
      return 0
    elif i==MX_PERVERTEX:
      return max(1, len(m.vertices))
    else:
      return 0

  def texcoord_binding(self, b:'byte'=0xffffffff):
    if b==0xffffffff: return (self.tbinding & (self.binding_mask >> 4))
    
    if b!=MX_UNBOUND and b!=MX_PERVERTEX:
      raise Exception("Illegal texture coordinate binding.")

    MxStdModel.binding_size(self, b)
    if self.tcoords:
      del self.tcoords[:] # .reset()
    else:
      self.tcoords = [] # MxDynBlock(size, TexCoord2f())

    self.tbinding = b

  bindings = [ "unbound", "face", "vertex" ]

  def binding_name(self, b):
    return None if b > MX_MAX_BINDING else MxStdModel.bindings[b]


  def compute_face_normal(self, f:MxFace, will_unitize=True):
    v1 = self.vertices[f.v0]     

    a = Point3d(self.vertices[f.v1]).sub(v1) # mixmops.mxv_sub(a, v2, v1, 3)
    b = Point3d(self.vertices[f.v2]).sub(v1) # mixmops.mxv_sub(b, v3, v1, 3)
    n = a.cross(b)       # mixmops.mxv_cross3(n, a, b)

    if will_unitize:
      n.normalize() # mixmops.mxv_unitize(n, 3)

    return n
  
  def compute_face_area(self, f):
    return 0.5 * self.compute_face_normal(f, False).norme()



  #///////////////////////////////////////////////////////////////////////
  #/ Contraction and related operations
  #/
  def compact_vertices(self):
    newID = 0 # /*MxVertexID*/int newID = 0

    for oldID,vtx in enumerate(self.vertices):
      #if self.vertex_is_valid(oldID):
      if vtx.tag & MX_VALID_FLAG:
        if newID!=oldID:
          self.vertices[newID].set(vtx)

          if self.texcoord_binding()==MX_PERVERTEX:
            self.tcoords[newID].set(self.tcoords[oldID])

          # Because we'll be freeing the link lists for the
          # old vertices, we actually have to swap values instead
          # of the simple copying in the block above.
          t = self.face_links[newID][:]
          self.face_links[newID][:] = self.face_links[oldID][:]
          self.face_links[oldID][:] = t[:] # .deepSet(t)

          # self.vertex_mark_valid(newID)
          self.vertices[newID].tag |= MX_VALID_FLAG

          for fid in self.face_links[newID]:
            self.faces[fid].remap_vertex(oldID, newID)
        newID+=1

    del self.face_links[newID:]
    del self.vertices[newID:]
    if self.texcoord_binding() == MX_PERVERTEX:
      del self.tcoords[newID:]
  
    

  def mark_neighborhood(self, vid, mark):
    for fi in self.face_links[vid]:
      self.faces[fi].mark = mark

  def collect_unmarked_neighbors(self, vid, faces:'list(int))'):
    for fid in self.face_links[vid]:
      if self.faces[fid].mark==0:
        faces.append(fid)
        self.faces[fid].mark = 1

  #public void collect_edge_neighbors(/* MxVertexID */int v1, /* MxVertexID */int v2, MxFaceList faces)
  def collect_edge_neighbors(self, v1, v2, faces):
    self.mark_neighborhood(v1, 1)
    self.mark_neighborhood(v2, 0)
    self.collect_unmarked_neighbors(v1, faces)

  def mark_neighborhood_delta(self, vid, delta):
    for fid in self.face_links[vid]:
      self.faces[fid].mark = 0xff & (self.faces[fid].mark + delta)

  #void partition_marked_neighbors(/*MxVertexID **/int v, int pivot, MxFaceList lo, MxFaceList hi)
  def partition_marked_neighbors(self, v, pivot, lo, hi):
    for fid in self.face_links[v]:
      if self.faces[fid].mark!=0:
        if self.faces[fid].mark < pivot:
          lo.append(fid)
        else:
          hi.append(fid)
        self.faces[fid].mark = 0

  #void mark_corners(MxFaceList faces, byte mark)
  def mark_corners(self, faces, mark):
    for fa in faces:
      self.vertices[self.faces[fa].v0].mark = mark
      self.vertices[self.faces[fa].v1].mark = mark
      self.vertices[self.faces[fa].v2].mark = mark
    
  #void collect_unmarked_corners(MxFaceList faces, MxVertexList verts)
  def collect_unmarked_corners(self, faces, verts):
    for fa in faces:
      for v in (self.faces[fa].v0,self.faces[fa].v1,self.faces[fa].v2):
        if self.vertices[v].mark==0:
          verts.append(v)
          self.vertices[v].mark = 1

  #public void collect_vertex_star(/* MxVertexID */int v, MxVertexList verts)
  def collect_vertex_star(self, v, verts):
    N = self.face_links[v] # list of int
    self.mark_corners(N, 0)
    self.vertices[v].mark = 1
    self.collect_unmarked_corners(N, verts)

  #void compute_contraction(/*MxVertexID*/int v1, /*MxVertexID*/int v2, MxPairContraction conx, float[] vnew /* = null */)
  def compute_contraction(self, v1, v2, conx, vnew=None):
    conx.v1 = v1
    conx.v2 = v2

    if vnew:
      #mixmops.mxv_sub(conx.dv1, vnew, vertex(v1).pos, 3)
      conx.dv1.sub(vnew, self.vetices[v1])
      conx.dv2.sub(vnew, self.vetices[v2])
    else:
      conx.dv1.set(0.0, 0.0, 0.0)
      conx.dv2.set(0.0, 0.0, 0.0)

    del conx.delta_faces[:] # .reset()
    del conx.dead_faces[:]

    # Mark the neighborhood of (v1,v2) such that each face is
    # tagged with the number of times the vertices v1,v2 occur
    # in it.  Possible values are 1 or 2.
    self.mark_neighborhood(v2, 0)
    self.mark_neighborhood(v1, 1)
    self.mark_neighborhood_delta(v2, 1)

    # Now partition the neighborhood of (v1,v2) into those faces
    # which degenerate during contraction and those which are merely
    # reshaped.
    self.partition_marked_neighbors(v1, 2, conx.delta_faces, conx.dead_faces)
    conx.delta_pivot = len(conx.delta_faces)
    self.partition_marked_neighbors(v2, 2, conx.delta_faces, conx.dead_faces)

  def unlink_face(self, fid):
    f = self.faces[fid]
    # self.face_mark_invalid(fid)
    f.tag &= ~MX_VALID_FLAG

    for flnk in (self.face_links[f.v0], self.face_links[f.v1], self.face_links[f.v2]):
      try:
        flnk.remove(fid)
      except ValueError:
        pass
    
      

  def apply_contraction(self, conx):
    v1 = conx.v1
    v2 = conx.v2

    # Move v1 to new position
    self.vertices[v1].add(conx.dv1)

    # Remove dead faces
    for fid in conx.dead_faces:
      self.unlink_face(fid)

    # Update changed faces
    for fid in conx.delta_faces[conx.delta_pivot:]:
      self.faces[fid].remap_vertex(v2, v1)
      self.face_links[v1].append(fid)

    # Kill v2
    self.vertices[v2].mark &= ~MX_VALID_FLAG
    del self.face_links[v2][:] # .reset()
  
 


  def read(self, geom:'WaveGeom', protEdges=False, protMat:'string'=None, lstProtFaces = None):
    ''' Convert a WaveGeom to an MxStdModel.
    
    Parameters:
    -----------
     protEdges: Boolean, optional, default False
      Indicate to the algorithm to freeze edges vertex.
      (It provides the means to protect interfaces between figures parts)
      
    protMat: str, optional, default None
      Indicate to the algorithm ot protect the faces with the given
      material name.
      
    lstProtFaces: iterable, optional, default None
      Indicate to the algorithm ot protect a given set of faces.
      Faces' numbers are relative to the input geometry (of course)
    
    '''
    # Copy Material list
    self.lstMat = geom.getMaterialList().copy()
    self.lstGrpName = [ g.getName() for g in geom.getGroups() ]
    
    # Copy all vertex
    for p in geom.coordList: # Avoid the getter who creates a copy
      self.add_vertex(p.x, p.y, p.z)

    # Create a tex coord mapping table
    # SMF format supporte one texture coord per vertex
    hasTexture = (geom.texList!=None) and (len(geom.texList) > 0)
    
    self.texcoord_binding(MX_PERVERTEX if hasTexture else MX_UNBOUND)

    if protMat:
      try:
        protMatIdx = geom.getMaterialList().index(protMat)
      except ValueError:
        logging.warning("Material[%s] is missing", protMat )

    if hasTexture:
      mapVT = [-1] * geom.getCoordListLength()
      try:
        for grp in geom.getGroups():
          vertIdxTbl = grp.vertIdx
          tvertIdxTbl = [-1] * len(grp.vertIdx) # grp.tvertIdx
          
          # I must admit that this kind of Vertex --> Texture mapping is FALSE!!
          for faceno, fao in enumerate(grp.matIdx):
            if fao.fAttr & C_FACE_TEXT:
              si = grp.getFaceStartIdx(faceno)
              ei = grp.getFaceLastIdx(faceno)
              tvertIdxTbl[si:ei] = fao.tvertIdx

          for i,v in enumerate(vertIdxTbl):
            mapVT[v] = tvertIdxTbl[i]
          
            
      except IndexError:
        logging.warning("Geom[%s] has bad texture index", geom.getName())
        hasTexture = False
        
      # Write out the texture coordinates
      nberror = 0
      for i in range(0, geom.getCoordListLength()):
        if mapVT[i]>=0:
          t = geom.texList[mapVT[i]]
        else:
          t = TexCoord2f(0, 0)
          nberror+=1
        self.add_texcoord(t.x, t.y)

    # Write the face of each group
    for grpNo, grp in enumerate(geom.getGroups()):
      
      # Keep group names
      self.lstGrpName.append(grp.getName())
      
      # Compute group edges (if required) and protect vertices   
      if protEdges:
        edges = grp.findEdges()
        for ed in edges:
          self.protvert.add(ed.idx0)
          self.protvert.add(ed.idx1)
          
      vertIdxTbl = grp.vertIdx
      for noface in range(0, grp.getNbFace()):
        startIdx = grp.getFaceStartIdx(noface)
        lastIdx = grp.getFaceLastIdx(noface)
        argc = lastIdx - startIdx
        matIdx = grp.getMatIdx(noface)

        isprotected = (protMat and protMatIdx==matIdx) or \
          (lstProtFaces and (noface in lstProtFaces))

        v0 = vertIdxTbl[startIdx]
        v1 = vertIdxTbl[startIdx+1]
        v2 = vertIdxTbl[startIdx+2]

        if argc==3:
          self.add_face(v0, v1, v2, isprotected, matIdx, grpNo)
        elif argc==4:
          v3 = vertIdxTbl[startIdx+3]          

          e0 = Point3d(self.vertices[v1]).sub(self.vertices[v0]).normalize()
          e1 = Point3d(self.vertices[v2]).sub(self.vertices[v1]).normalize()
          e2 = Point3d(self.vertices[v3]).sub(self.vertices[v2]).normalize()
          e3 = Point3d(self.vertices[v0]).sub(self.vertices[v3]).normalize()

          a_02 = (1 - e0.dot(e3)) + (1 - e1.dot(e2))
          a_13 = (1 - e1.dot(e0)) + (1 - e3.dot(e2))
          # Comparison with Java Algo : Accuracy changes
          if a_02 <= a_13:
            self.add_face(v0, v1, v2, isprotected, matIdx, grpNo)
            self.add_face(v0, v2, v3, isprotected, matIdx, grpNo)
          else:
            self.add_face(v0, v1, v3, isprotected, matIdx, grpNo)
            self.add_face(v1, v2, v3, isprotected, matIdx, grpNo)
        else:
          # print('Input polygon #{:d} has {:d} sides. Triangularize it.'.format(noface, argc))
          
          # Compute the  isobarycentre
          isoG = Point3d()
          t= TexCoord2f()
          for vno in vertIdxTbl[startIdx:lastIdx]:
            isoG.add(self.vertices[vno])
            if hasTexture:
              t.add(geom.texList[mapVT[vno]])
          gno = len(self.vertices)
          isoG.scale(1.0/float(argc))
          t.scale(1.0/float(argc))
          self.add_vertex(isoG.x, isoG.y, isoG.z)
            
          if hasTexture:
            self.add_texcoord(t.x, t.y)
              
          for i in range(startIdx, lastIdx-1):
            self.add_face(gno, vertIdxTbl[i], vertIdxTbl[i+1], isprotected, matIdx, grpNo)
            
          self.add_face(gno, vertIdxTbl[lastIdx-1], vertIdxTbl[startIdx], isprotected, matIdx, grpNo)

    logging.info("Geom[%s] with %d polygons has %d triangles", geom.getName(), geom.getNbFace(), len(self.faces))
    return self


  def createWaveGeom(self, restoreGrp=False, restoreQuad=False):
    ''' Convert a MxStdModel into a WaveGeom.
    
    Parameters
    ----------
    restoreQuad: Boolean, optional, default False
      Try to merge triangles to rebuild 'quadrangle' faces
      
    restoreGrp: Boolean, optional, default False
      Preserve (and recreate) the groups in the result geometry.
      
    Returns
    -------
      WaveGeom
        A new WaveGeom
    '''      
    # Create the new (output) WaveGeom
    nwg = WaveGeom()
    nwg.lstMat = self.lstMat.copy()

    # Rebuild all groups (to preserve group indexes)
    for grpname in self.lstGrpName if restoreGrp else [ 'decimate',]:
      nwg.createGeomGroup(grpname)

    # Create new Point3d, because they've been modified with .mark and .tag
    nwg.coordList += [ Point3d(v) for v in self.vertices ]

    hasTexture = (self.texcoord_binding()!=MX_UNBOUND)
    if hasTexture:
      nwg.texList += self.tcoords

    prevFace = None
    prefGrpNo = -1
    curGroup = nwg.getGroups()[0]
    
    for fx in self.faces: # range(0, self.face_count()):
      
      if restoreGrp and (prefGrpNo!=fx.grpNo):
        curGroup  = nwg.getGroups()[fx.grpNo]
        prefGrpNo = fx.grpNo
        prevFace  = None  
      
      if fx.tag & MX_VALID_FLAG:
        # Add face to current groups
        if restoreQuad and prevFace and (curGroup.curMatIdx == fx.matIdx):
          lprevEdge = ( (prevFace[2], prevFace[1]), (prevFace[0], prevFace[2]), (prevFace[1], prevFace[0]) )
           
          try:
            keepNo = lprevEdge.index( (fx.v0, fx.v1) )            
          except ValueError:
            keepNo = -1
            
          if keepNo>0: # Keep pv2, remove edge v0, v1
            # Remove previous face (the last one)
            curGroup.removeFace(faceno=curGroup.getNbFace()-1)
            
            # Previous face as a common edge - Merge along this edge
            l = [fx.v0, prevFace[keepNo], fx.v1, fx.v2]
            curGroup.curMatIdx = fx.matIdx
            curGroup.addFace(l, l if hasTexture else [], [])
            prevFace = None
            continue

          try:
            keepNo = lprevEdge.index( (fx.v1, fx.v2) )            
          except ValueError:
            keepNo = -1
            
          if keepNo>0: # Keep pv0, remove edge v1, v2
            # Remove previous face (the last one)
            curGroup.removeFace(faceno=curGroup.getNbFace()-1)
            
            # Previous face as a common edge - Merge along this edge
            l = [fx.v0, fx.v1, prevFace[keepNo], fx.v2]
            curGroup.curMatIdx = fx.matIdx
            curGroup.addFace(l, l if hasTexture else [], [])
            prevFace = None
            continue

          try:
            keepNo = lprevEdge.index( (fx.v2, fx.v0) )            
          except ValueError:
            keepNo = -1
            
          if keepNo>0: # Keep pv1, remove edge v2, v0
            # Remove previous face (the last one)
            curGroup.removeFace(faceno=curGroup.getNbFace()-1)
            
            # Previous face as a common edge - Merge along this edge
            l = [fx.v0, fx.v1, fx.v2, prevFace[keepNo]]
            curGroup.curMatIdx = fx.matIdx
            curGroup.addFace(l, l if hasTexture else [], [])
            prevFace = None
            continue

        #else:
        # No Quadrangle restoration OR
        # Previous face as a different material OR
        # Not 'continue' (ie. faces are not sharing any edge)
        prevFace = [fx.v0, fx.v1, fx.v2]
        curGroup.curMatIdx = fx.matIdx
        curGroup.addFace(prevFace, prevFace if hasTexture else [], [])

    # Close Groups
    nwg.optimizeGroups(cleaning=False)

    return nwg









